<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2011 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

*/


abstract class Storm_Model_Abstract implements Storm_Model_Interface
{

  protected static array $_loaders = [];
  protected static array $_method_prefixes = ['get' => 3,
                                              'set' => 3,
                                              'add' => 3,
                                              'remove' => 6,
                                              'has' => 3,
                                              'numberOf' => 8];

  protected ?Storm_Model_State $_model_state = null;

  protected string $_table_primary = 'id';

  protected array $_attributes = ['id' => null];

  protected array $_attributes_in_db = [];

  protected array $_has_many_attributes_in_db = [];

  protected array $_has_many_attributes = [];

  /**
   * Defines me -> * dependents relationship
   * Attributes:
   *   model:  dependent class name
   *   role: the name of the accessor (also used for default id mapping in db)
   *   order: order statetement while loading models
   *   through: this relation relies on another one
   *   unique: if true, a dependent could not appear twice
   *   dependents: if 'delete', then deleting me will delete dependents
   *   scope: an array of field / value pair that filters data
   *   limit: when specified, add a limit for fetching descendents
   *
   * Ex:
   *  protected $_has_many = array('sessions' => array(
   *                                                 'model' => 'Class_SessionFormation',
   *                                                 'role' => 'formation',
   *                                                 'dependents' => 'delete',
   *                                                 'order' => 'date_debut desc',
   *                                                 'scope' => array('type' => 2)),
   *
   *                               'session_formation_inscriptions' => array('through' => 'sessions',
   *                                                                         'unique' => 'true'),
   *
   *                               'stagiaires' => array('through' => 'session_formation_inscriptions'));
   *
   * Relationships definition can be overrided at runtime by writing a method named descriptionOfFieldName.
   * This can be used to have dynamic relationships between objects.
   *
   * ex:
   *
   * protected $_has_many = [ 'animals' => [] ];
   *
   *  public function descriptionOfAnimals() {
   *      return $this->iWantCats()
   *                  ? [ 'model' => 'Cat', 'role' => 'owner' ]
   *                  : [ 'model' => 'Dog', 'role' => 'owner' ];
   *  }
   */
  protected array $_has_many = [];

  protected array $_belongs_to_attributes = [];

  /**
   * Defines me -> * dependents relationship
   * Attributes:
   *   model:  dependent class name
   *   referenced_in: field name used to reference dependent id
   *   trough: this relation relies on another one
   * protected $_belongs_to = array('librairy' => array('model' => 'Librairy',
   *                                                    'referenced_in' => 'id_library'),
   *
   *                                'zone' => array('through' => 'librairy'));
   */
  protected array $_belongs_to = [];

  /**
   * Store the list of errors found by validate() method.
   * See also check($condition, $error)
   */
  protected array $_errors = [];

  /**
   * Default values for attributes of a new instance
   * Should be defined in subclasses like:
   * $_default_values = array('title' => 'new article', 'content' => '')
   */
  protected array $_default_attribute_values = [];

  /**
   * By default, use autoincrement ids generated by the SGBD. When $_fixed_id = true,
   * does not take id from SGBD.
   * @var boolean
   */
  protected bool $_fixed_id = false;

  /**
   * Generic mean to handle associations (in order to extent and replace has_many and belongs_to)
   * Collection of Storm_Model_Association_Abstract subclasses
   */
  protected Storm_Model_Associations $_associations;

  protected static function _buildLoaderFor(string $class): Storm_Model_Loader
  {
    $reflection = new ReflectionClass($class);
    $class_vars = $reflection->getdefaultProperties();

    if (isset($class_vars['_loader_class'])) {
      $loader_class = $class_vars['_loader_class'];

      return new $loader_class($class);
    }

    return new Storm_Model_Loader($class);
  }

  public static function getClassVar(string $var)
  {
    $reflection = new ReflectionClass(static::class);
    $class_vars = $reflection->getdefaultProperties();

    return $class_vars[$var];
  }

  public static function getLoader()
  {
    return static::getLoaderFor(static::class);
  }

  /**
   * Forward static calls to the Loader so instead of writing:
   *    My_Class::getLoader()->foo()
   * we can write:
   *    My_Class::()
   * @param string $name method name
   * @param array $args
   * @return mixed
   */
  public static function __callStatic($name, $args)
  {
    return call_user_func_array([static::getLoader(), $name], $args);
  }

  public static function getLoaderFor(string $class)
  {
    $parent_class = get_parent_class($class);
    if ( ! $parent_class || ($parent_class === Storm_Model_Abstract::class))
      return (static::$_loaders[$class]
              ?? static::setLoaderFor($class, static::_buildLoaderFor($class)));

    return ($parent_loader = (static::$_loaders[$parent_class] ?? null))
      ? (static::$_loaders[$class] = $parent_loader->setInstanceClass($class))
      : (static::setLoaderFor($class, static::_buildLoaderFor($parent_class), $parent_class)
         ->setInstanceClass($class));
  }

  public static function setLoaderFor(string $class,
                                      $loader,
                                      string $parent_class = '')
  {
    if ($parent_class)
      static::$_loaders [$parent_class] = $loader;

    return static::$_loaders [$class] = $loader;
  }

  public static function unsetLoaders(): void
  {
    static::$_loaders = [];
  }

  public static function getLoaders(): array
  {
    return static::$_loaders;
  }

  public function __construct()
  {
    $this->_associations = new Storm_Model_Associations;
    $this->describeAssociationsOn($this->_associations);
  }

  public function describeAssociationsOn(Storm_Model_Associations $associations): void {}

  public function hasBelongsToRelashionshipWith(string $field): bool
  {
    return array_key_exists($field, $this->_belongs_to);
  }

  public function cache(): self
  {
    static::getLoader()->cacheInstance($this);

    return $this;
  }

  public function getClassName(): string
  {
    return get_class($this);
  }

  protected function _getEvents(): Storm_Events
  {
    return Storm_Events::getInstance();
  }

  public function assertSave(bool $force_primary_key = false): bool
  {
    if ( ! $this->save($force_primary_key))
      throw new Storm_Model_Exception('Can\'t save '
                                      . get_class($this)
                                      . ': ' . implode(', ', $this->getErrors()));

    return true;
  }

  public function save(?bool $force_primary_key = false): bool
  {
    $this->_updateNullBelongsToIdFieldsFromDependents();

    return $this->isValid()
      ? $this->saveWithoutValidation(($force_primary_key ?? false), true)
      : $this->revertAttributes();
  }

  public function revertAttributes(): bool
  {
    if ( ! $this->isNew() && $this->_attributes_in_db)
      $this->_attributes = $this->_attributes_in_db;

    return false;
  }

  public function changedId(): mixed
  {
    return (($old_id = ($this->_attributes_in_db['id'] ?? null))
            && ($old_id != ($this->_attributes['id'] ?? null)))
      ? $old_id
      : null;
  }

  public function saveWithoutValidation(bool $force_primary_key = false,
                                        bool $from_save = false): bool
  {
    $is_changed_id = (null !== $this->changedId());

    $this->beforeSave();
    $this->_associations->saveBefore($this);

    $this->_updateNullBelongsToIdFieldsFromDependents()
         ->_belongsToDiff($is_changed_id);

    $has_manys = [];
    if ($is_changed_id)
      foreach ($this->_has_many as $field => $has_many)
        $has_manys [$field] = $this->_getDependents($field);

    $force_primary_key
      ? $this->getLoader()->saveForcePrimaryKey($this)
      : $this->getLoader()->save($this);

    if ($is_changed_id)
      foreach ($has_manys as $field => $has_many)
        $this->_setDependents($field, $has_many);

    $this->cache()
         ->_saveDependencies()
         ->afterSave();

    if ($from_save)
      $this->_getEvents()->notify(new Storm_Event_Save($this));

    $this->_attributes_in_db = $this->_attributes;

    return true;
  }

  protected function _updateNullBelongsToIdFieldsFromDependents(): self
  {
    $this->_belongs_to_attributes = array_filter($this->_belongs_to_attributes);

    foreach ($this->_belongs_to_attributes as $field => $dependent) {
      $id_field = $this->getIdFieldForDependent($field);

      if (null === $this->_get($id_field))
        $this->_setAttributes($id_field, $dependent->getId());
    }

    return $this;
  }

  public function beforeSave(): self
  {
    return $this;
  }

  public function afterSave(): self
  {
    return $this;
  }

  public function beforeDelete(): self
  {
    return $this;
  }

  public function afterDelete(): self
  {
    return $this;
  }

  /**
   * Is this model valid for saving
   *
   * You can notify error by using addError or check methods
   *
   * @see Storm_Model_Abstract::addError
   * @see Storm_Model_Abstract::check
   *
   * @example
   * public function validate() {
   *    $this->check($this->getRole() < 10, 'role should not exceed 10');
   * }
   *
   */
  public function validate(): self
  {
    return $this;
  }

  /**
   * Try to validate the model. If errors found, return false
   * @see Storm_Model_Abstract::validate
   */
  public function isValid(): bool
  {
    $this->_errors = [];
    $this->validate();

    return ! $this->hasErrors();
  }

  /**
   * Return true if contains some errors (do not try to validate)
   * @see Storm_Model_Abstract::validate
   */
  public function hasErrors(): bool
  {
    return count($this->getErrors()) > 0;
  }

  public function getErrors(): array
  {
    return $this->_errors;

  }

  public function addAttributeError(string $attribute, string $error): self
  {
    $this->_errors[$attribute] = $error;

    return $this;
  }

  public function addError(string $error): self
  {
    $this->_errors[] = $error;

    return $this;
  }

  public function check(bool $condition, string $error): self
  {
    if ( ! $condition)
      $this->addError($error);

    return $this;
  }

  public function checkAttribute(string $attribute,
                                 bool $condition,
                                 string $error): self
  {
    if ( ! $condition)
      $this->addAttributeError($attribute, $error);

    return $this;
  }

  /**
   * Validate an attribute with a validator
   */
  public function validateAttribute(string $name,
                                    string $validator_class,
                                    ?string $message = null): self
  {
    $validator = new $validator_class;
    $valid = $validator->isValid($this->_get($name));
    if ($message)
      return $this->checkAttribute($name, $valid, $message);

    foreach ($validator->getMessages() as $message)
      $this->checkAttribute($name, $valid, $message);

    return $this;
  }

  /**
   * Return an associative array with attribute
   * name as $key and its value.
   *
   * Used by Loader while saving in order to build the
   * SQL query.
   */
  public function attributesToArray(): array
  {
    $attributes = [];

    $all_attributes = array_merge($this->_default_attribute_values,
                                  $this->_attributes);

    foreach ($all_attributes as $name => $value) {
      $method = 'get' . $this->attributeNameToAccessor($name);

      $attributes[$name] = method_exists($this, $method)
        ? $this->$method()
        : $value;
    }

    return $attributes;
  }

  public function getRawAttributes(): array
  {
    return array_merge($this->_default_attribute_values,
                       $this->_attributes);
  }

  public function getRawDbAttributes(): array
  {
    return $this->_attributes_in_db;
  }

  /**
   * Return an associative array with attribute
   * name as $key and its value.
   *
   * This method may be redefined in subclasses
   * in order to provide some coupling with Zend_Form::populate()
   */
  public function toArray(): array
  {
    return $this->attributesToArray();
  }

  public function isNew(): bool
  {
    return null === ($this->_attributes['id'] ?? null);
  }

  public function delete(): void
  {
    $this->forceDelete();
  }

  public function forceDelete(): self
  {
    $this->beforeDelete()
         ->_deleteDependents();

    if ($this->isNew() || $this->getLoader()->delete($this))
      $this->_doForBelongsTo()
           ->afterDelete()
           ->_getEvents()->notify(new Storm_Event_Delete($this));

    return $this;
  }

  protected function _belongsToDiff(bool $is_changed_id): self
  {
    if ($is_changed_id)
      return $this;

    foreach ($this->_belongs_to as $role => $belongs_to) {
      $class_name = $belongs_to['model'] ?? '';
      if ( ! $role || ! $class_name)
        continue;

      $id_field = $this->getIdFieldForDependent($role);

      $new_id = $this->_attributes[$id_field] ?? null;
      $old_id = $this->_attributes_in_db[$id_field] ?? null;

      if ($old_id === $new_id)
        continue;

      unset($this->_belongs_to_attributes[$role]);
      $instances = $class_name::getLoader()->getLoadedInstances(false);

      if ($old_dependent = ($instances[$old_id] ?? null))
        $old_dependent->_doForHasMany($role, $this);

      if ($dependent = ($instances[$new_id] ?? null))
        $dependent->_doForHasMany($role, $this);
    }

    return $this;
  }

  protected function _doForBelongsTo(): self
  {
    foreach ($this->_belongs_to as $role => $belongs_to)
      if ($role && ($dependent = $this->_getDependent($role, false)))
        $dependent->_doForHasMany($role, $this);

    return $this;
  }

  protected function _doForHasMany(string $role, Storm_Model_Abstract $belong): self
  {
    $class = $belong->getClassName();

    foreach ($this->_has_many as $many => $has_many)
      if ($role === ($has_many['role'] ?? '')
          && $class === ($has_many['model'] ?? ''))
        $this->_removeOnHasManysFor($many, $has_many);

    return $this;
  }

  protected function _removeOnHasManysFor(string $many, array $has_many): self
  {
    unset($this->_has_many_attributes[$many]);
    unset($this->_has_many_attributes_in_db[$many]);

    return $this->_resetHasManyCache($has_many);
  }

  protected function _resetHasManyCache(array $has_many): self
  {
    if (($query = $this->_queryLoadDependents($has_many))
        && $query->notLoadInCache()
        && ($loader = $this->_getLoaderForModel($has_many['model'] ?? '')))
      $loader->removeFromCachedInstance($query->uniqKey());

    return $this;
  }

  public function resetCacheAllDependents(): self
  {
    $this->_has_many_attributes = [];
    $this->_has_many_attributes_in_db = [];

    return $this->_doForBelongsTo();
  }

  public function setId(mixed $id): self
  {
    if ($this->_table_primary)
      $this->_setAttributes(strtolower($this->_table_primary), $id);

    return $this->_setAttributes('id', $id);
  }

  public function getIdFieldForDependent(string $field): string
  {
    return ($id_field = ($this->_belongs_to[$field]['referenced_in'] ?? ''))
      ? $id_field
      : ($field . '_id');
  }

  public function dependentsDeleteClassAndId(): array
  {
    $dependents_ids = [];

    foreach ($this->_has_many as $field => $relation)
      if (('delete' === ($relation['dependents'] ?? ''))
          && ($class_name = ($relation['model'] ?? ''))
          && ($id_field = (new $class_name)->idfieldBelongsFromClassname(static::class)))
        $dependents_ids [$class_name] = $id_field;

    return $dependents_ids;
  }

  public function idfieldBelongsFromClassname(string $class_name): string
  {
    foreach ($this->_belongs_to as $belongs)
      if ($class_name === ($belongs['model'] ?? ''))
        return $belongs['referenced_in'] ?? '';

    return '';
  }

  public function isAttributeEmpty(string $attribute): bool
  {
    return array_key_exists($attribute, $this->_attributes)
      ? empty($this->_attributes[$attribute] ?? '')
      : true;
  }

  protected function _deleteDependents(): self
  {
    foreach ($this->hasManyRelationships() as $field => $relation) {
      if ( ! array_key_exists('dependents', $relation))
        continue;

      if ($relation['dependents'] != 'delete')
        continue;

      $dependents = $this->_getDependents($field);

      foreach ($dependents as $object) {
        $this->_removeDependent($field, $object);
        $object->delete();
      }
    }

    $this->_associations->delete($this);

    return $this;
  }

  protected function _saveDependencies(): self
  {
    foreach (array_keys($this->_has_many_attributes) as $field)
      $this->_saveDependents($field);

    $this->_associations->save($this);

    return $this;
  }

  protected function _saveDependents(string $field): self
  {
    if (array_key_exists('through', $this->descriptionOf($field)))
      return $this;

    if ( ! array_key_exists($field, $this->_has_many_attributes_in_db))
      $this->_has_many_attributes_in_db [$field] = $this->_getDependentsFromLoader($field);

    $has_many_attributes = (array) ($this->_has_many_attributes[$field] ?? []);
    $has_many_attributes_in_db = (array) ($this->_has_many_attributes_in_db[$field] ?? []);
    $dependents_to_delete = array_diff($has_many_attributes_in_db,
                                       $has_many_attributes);

    foreach ($dependents_to_delete as $dependent)
      $dependent->delete();

    foreach ($has_many_attributes as $dependent)
      $dependent->save();

    if ($has_many_for_db = ($this->_has_many_attributes[$field] ?? []))
      $this->_has_many_attributes_in_db [$field] = $has_many_for_db;

    return $this->_resetHasManyCache(($dependents_to_delete || $has_many_attributes)
                                     ? $this->_has_many[$field]
                                     : []);
  }

  /**
   * Main purpose is to setup generic getters and setters:
   *
   * $car->getColor();
   * $car->setColor('red');
   * $car->addWheel($w = new Wheel());
   * $car->removeWheel($w);
   * $car->getWheels()   //return an array containing instances;
   * $car->setWheels(array(new Wheel(), new Wheel(), new Wheel(), new Wheel()))
   * $car->hasWheels()   //return true if attribute wheels not empty;
   *
   * @param string $method
   * @param array $args
   * @return Storm_Model_Abstract
   * @throws Exception
   */
  public function __call($method, $args) {
    return $this->_associations->handleCall($this, $method, $args);
  }

  protected function _methodMatch(string $method): ?array
  {
    foreach (static::$_method_prefixes as $prefix => $length)
      if (substr($method, 0, $length) == $prefix)
        return [1 => $prefix,
                2 => substr($method, $length)];

    return null;
  }

  /**
   * See __call
   *
   * ex:
   * $car->perform('getColor')
   * $car->perform('setColor', ['red'])
   */
  public function perform(string $method, array $args = []): mixed
  {
    if ( ! ($matches = $this->_methodMatch($method)))
      throw new Storm_Model_Exception('Tried to call unknown method '
                                      . get_class($this) . '::' . $method);

    $attribute = $this->_accessorToAttributeName($matches[2]);

    switch ($matches[1]) {
      case 'get':
        return $this->_get($attribute);
      case 'set':
        return $this->_set($attribute, $args[0]);
      case 'add':
        return $this->_addDependent($this->_pluralize($attribute), $args[0]);
      case 'remove':
        return $this->_removeDependent($this->_pluralize($attribute), $args[0]);
      case 'has':
        return $this->_has($attribute);
      case 'numberOf':
        return $this->_numberOf($attribute);
    }

    return $this;
  }

  public function descriptionOf(string $field): mixed
  {
    $method = 'descriptionOf' . $field;

    if (method_exists($this, $method))
      return $this->$method();

    if (isset($this->_has_many[$field]))
      return $this->_has_many[$field];

    return $this->_belongs_to[$field] ?? null;
  }

  public function hasManyRelationships(): array
  {
    $relations = $this->_has_many;

    foreach (array_keys($this->_has_many) as $field)
      $relations[$field] = $this->descriptionOf($field);

    return $relations;
  }

  /**
   * Return true if $attribute not empty
   */
  public function _has(string $attribute): bool
  {
    $dependent = $this->callGetterByAttributeName($attribute);

    return ! empty($dependent);
  }

  /**
   * Returns the number of dependents in a has_many relationship
   */
  protected function _numberOf(string $field): int
  {
    if ( ! ($map = $this->descriptionOf($field)))
      throw new Storm_Model_Exception(sprintf('Tried to call unknown method %s::numberOf%s',
                                              get_class($this),
                                              $this->attributeNameToAccessor($field)));

    if (isset($map['through']))
      return $this->_numberOfThrough($field, $map);

    if (array_key_exists($field, $this->_has_many_attributes))
      return count($this->_has_many_attributes[$field]);

    if (isset($this->_belongs_to[$field]))
      return $this->_getDependent($field) ? 1 : 0;

    return ($query = $this->_queryLoadDependents($map, true))
      ? $query->countAll()
      : 0;
  }

  protected function _numberOfThrough(string $field, array $map): int
  {
    $through_field = $map['through'];
    $is_unique = $map['unique'] ?? false;

    if ($is_unique)
      return $this->_numberOf($through_field);

    $dependents = $this->_get($through_field);
    if (isset($this->_belongs_to[$through_field]))
      $dependents = array_filter([$dependents]);

    return array_sum(array_map(fn($dependent) => $dependent
                               ->_numberOf($this->_getDependentFieldNameForInstance($field,
                                                                                    $dependent)),
                               $dependents));
  }

  /**
   * Update with a [name] => value formatted array
   */
  public function updateAttributes(array $datas): self
  {
    foreach ($datas as $name => $value) {
      $method = 'set' . $this->attributeNameToAccessor($name);
      $this->$method($value);
    }

    return $this;
  }

  /**
   * Set initial attribute values with a [name] => value formatted array
   * Used by Loader::newFromRow
   * The difference with updateAttributes is that it doesn't call
   * setAttributeName(...) magic.
   *
   * So we are sure that attributes are exactly those from db.
   */
  public function initializeAttributes(array $datas): self
  {
    $this->_attributes_in_db = $this->_attributes = array_merge($this->_attributes,
                                                                array_change_key_case($datas));

    return $this;
  }

  /**
   * @return boolean true when attribute has changed from initial value in db
   */
  public function hasChangedAttribute(?string $name): bool
  {
    if ($this->isNew()
        || null === $name
        || ! isset($this->_attributes[$name]))
      return false;

    if ( ! isset($this->_attributes_in_db[$name]))
      return true;

    return $this->_attributes_in_db[$name] != $this->_attributes[$name];
  }

  public function hasChange(): bool
  {
    foreach (array_keys($this->toArray()) as $name)
      if ($this->hasChangedAttribute($name))
        return true;

    return false;
  }

  /**
   * UserId -> user_id
   */
  protected function _accessorToAttributeName(string $accessor): string
  {
    return Storm_Inflector::underscorize($accessor);
  }

  /**
   * user_id -> UserId
   */
  public function attributeNameToAccessor(string $name): string
  {
    return Storm_Inflector::camelize($name);
  }

  public function isAttributeExists(string $field): bool
  {
    return
      array_key_exists($field, $this->_attributes)
      || array_key_exists($field, $this->_has_many)
      || array_key_exists($field, $this->_belongs_to)
      || $this->hasDefaultValueForAttribute($field);
  }

  public function getId(): mixed
  {
    return $this->_attributes['id'];
  }

  /**
   * Return the value of attribute $field. If cannot find it, raise Exception.
   * @throws Storm_Model_Exception
   */
  protected function _get(string $field): mixed
  {
    if (isset($this->_attributes[$field]) || array_key_exists($field, $this->_attributes))
      return $this->_attributes[$field];

    if (isset($this->_has_many[$field]) || array_key_exists($field, $this->_has_many))
      return $this->_getDependents($field);

    if (isset($this->_belongs_to[$field]) || array_key_exists($field, $this->_belongs_to))
      return $this->_getDependent($field);

    if ($this->hasDefaultValueForAttribute($field))
      return $this->getDefaultValueForAttribute($field);

    throw new Storm_Model_Exception(sprintf('Tried to call unknown method %s::get%s',
                                            get_class($this),
                                            $this->attributeNameToAccessor($field)));
  }

  /**
   * Create compatibility with accessor like $miles->instrument.
   * ex:
   *    $miles->setInstrument("trumpet");
   *    assert("trumpet" === $miles->getTrumpet())
   *    assert("trumpet" === $miles->trumpet)
   *
   * If attribute not defined, returns null for compatibility purpose
   * with legacy code - Remember STORM is intended to get
   * easily into messy crappy code. Sorry.
   *
   *     assert(null ===  $miles->zork)
   */
  public function __get(string $field): mixed
  {
    return ($this->_attributes[strtolower($field)]
            ?? $this->_attributes[strtoupper($field)]
            ?? null);
  }

  /**
   * Tells whether a default value is defined for an attribute
   */
  public function hasDefaultValueForAttribute(string $field): bool
  {
    return array_key_exists($field, $this->_default_attribute_values);
  }

  /**
   * Return the default value for given attribute
   */
  public function getDefaultValueForAttribute(string $field): mixed
  {
    return $this->_default_attribute_values[$field];
  }

  public function __set(string $field, mixed $value)
  {
    $method = 'set' . strtolower($field);

    return $this->$method($value);
  }

  /**
   * If has_many relationship specifies unique => true and $value in $dependents,
   * then the constraint is violated.
   */
  protected function _isUniqueConstraintViolated(array $relationship,
                                                 $dependents,
                                                 Storm_Model_Abstract $value): bool
  {
    $is_unique = true === ($relationship['unique'] ?? true);
    if ( ! $is_unique)
      return false;

    if ($value->isNew())
      return false;

    foreach ($dependents as $dependent)
      if ($dependent->getId() == $value->getId())
        return true;

    return false;
  }

  protected function _addDependent(string $field, mixed $value): self
  {
    if ( ! $value)
      return $this;

    if ( ! ($map = $this->descriptionOf($field)))
      throw new Storm_Model_Exception(sprintf('Tried to call unknown method %s::add%s',
                                              get_class($this),
                                              $this->attributeNameToAccessor($this->_singularize($field))));

    if (isset($map['through']))
      return $this->_addDependentThrough($map, $field, $value);

    $dependents = $this->_get($field);
    if ($this->_isUniqueConstraintViolated($map, $dependents, $value))
      return $this;

    $dependents[] = $value;
    $this->_set($field, $dependents);

    $value->_set($map['role'], $this);

    return $this;
  }

  protected function _addDependentThrough(array $map,
                                          string $field,
                                          mixed $value): self
  {
    $through_field = $map['through'];

    if (isset($this->_belongs_to[$through_field]))
      return $this->_getDependent($through_field)->callAdderByAttributeName($field, $value);

    if ($this->_isUniqueConstraintViolated($map, $this->_get($field), $value))
      return $this;

    $through_field_map = $this->descriptionOf($through_field);
    $role = $through_field_map['role'];
    $model = $through_field_map['model'];

    $intermediate = new $model;
    $intermediate->_set($role, $this);
    $intermediate->_set($this->_singularize($field), $value);

    return $this->_addDependent($through_field, $intermediate);
  }

  protected function _findDependentOfMeAndValueThrough(string $field, Storm_Model_Abstract $value, string $through_field): ?Storm_Model_Abstract
  {
    $field_getter = 'get' . $this->attributeNameToAccessor($this->_singularize($field));
    $dependents = $this->_get($through_field);

    foreach ($dependents as $dependent) {
      if ( ! isset($get_id_method)) {
        $map = $this->descriptionOf($through_field);
        $get_id_method = 'get' . $this->attributeNameToAccessor($dependent->getIdFieldForDependent($map['role']));
      }

      if ($dependent->$get_id_method() !== $this->getId())
        continue;

      if (null == ($other_value = $dependent->$field_getter()))
        continue;

      if ($value->getId() == $other_value->getId())
        return $dependent;
    }

    return null;
  }

  protected function _removeDependentThrough(string $field,
                                             Storm_Model_Abstract $value,
                                             string $through_field): self
  {
    if ($dependent_to_remove = $this->_findDependentOfMeAndValueThrough($field,
                                                                        $value,
                                                                        $through_field))
      $this->_removeDependent($through_field, $dependent_to_remove);

    return $this;
  }

  protected function _removeDependent(string $field,
                                      ?Storm_Model_Abstract $value): self
  {
    if ( ! $value)
      return $this;

    if ( ! array_key_exists($field, $this->_has_many))
      throw new Storm_Model_Exception('Tried to call unknown method '
                                      . get_class($this)
                                      . '::remove' . $field);

    $map = $this->descriptionOf($field);
    return isset($map['through'])
      ? $this->_removeDependentThrough($field, $value, $map['through'])
      : $this->_setDependents($field, $this->_keepSameDependents($this->_getDependents($field),
                                                                 $value));
  }

  protected function _keepSameDependents(mixed $dependents,
                                         Storm_Model_Abstract $value): mixed
  {
    $callback = fn($dependent) => ! $this->_isSameDependent($dependent, $value);

    if (is_array($dependents))
      return array_filter($dependents, $callback);

    if ($dependents instanceof Storm_Collection)
      return $dependents->select($callback);

    return $dependents;
  }

  protected function _isSameDependent(Storm_Model_Abstract $dependent,
                                      Storm_Model_Abstract $value): bool
  {
    return $value->isNew()
      ? ($dependent->getRawAttributes() == $value->getRawAttributes())
      : ($dependent->getId() == $value->getId());
  }

  protected function _getDependents(string $field): mixed
  {
    if (isset($this->descriptionOf($field)['through']))
      return $this->_getDependentsFromLoader($field);

    if (isset($this->_has_many_attributes[$field]))
      return $this->_has_many_attributes[$field];

    $this->_storeDependentsForField($dependents = $this->_getDependentsFromLoader($field),
                                    $field);
    return $dependents;
  }

  public function _storeDependentsForField(mixed $dependents, string $field): self
  {
    $this->_has_many_attributes[$field] = $dependents;

    return $this;
  }

  protected function _getDependentsOfFieldThrough(string $field,
                                                  string $through_field): array
  {
    if (isset($this->_belongs_to[$through_field]))
      return ($dependent = $this->_getDependent($through_field))
        ? $dependent->_get($field)
        : [];

    return $this->_getDependentFieldForInstances($field,
                                                 $this->_getDependents($through_field));
  }

  public function _getDependentFieldForInstances(string $field,
                                                 array $instances): array
  {
    $dependents = [];

    $singularized_field = $this->_singularize($field);
    $getManyMethod = 'get' . ucfirst($field);
    $getOneMethod = 'get' . ucfirst($singularized_field);

    foreach ($instances as $instance) {
      if ($instance->hasBelongsToRelashionshipWith($singularized_field))
        $dependents [] = $instance->$getOneMethod();
      else
        $dependents = array_merge($dependents, $instance->$getManyMethod());
    }

    return array_filter($dependents);
  }

  public function _getDependentFieldNameForInstance(string $field, Storm_Model_Abstract $instance): string
  {
    $singularized_field = $this->_singularize($field);

    return $instance->hasBelongsToRelashionshipWith($singularized_field)
      ? $singularized_field
      : $field;
  }

  protected function _getLoaderForModel(string $model)
  {
    return call_user_func([$model, 'getLoader']);
  }

  protected function _getDependentsFromLoader(string $field): mixed
  {
    $map = $this->descriptionOf($field);

    if (isset($map['through']))
      return $this->_wrapDependentsWithDefinedClass($field,
                                                    $this->_getDependentsOfFieldThrough($field, $map['through']));

    return $this->_wrapDependentsWithDefinedClass($field,
                                                  ($query = $this->_queryLoadDependents($map))
                                                  ? array_filter($query->fetchAll(true))
                                                  : []);
  }

  protected function _queryLoadDependents(array $map,
                                          bool $for_count = false): ?Storm_Query
  {
    if ( ! $map || $this->isNew())
      return null;

    $loader = $this->_getLoaderForModel($map['model']);
    $id_field = $loader->getIdFieldForDependent($map['role']);

    $id = $this->changedId() ?? $this->_attributes['id'] ?? null;
    if (null === $id)
      return null;

    $query = $loader->query()->eq($id_field, $id);

    $this->_addScopeForQueryDependent($query, $map);

    if ( ! $for_count)
      $this->_addOrderForQueryDependent($query, $map)
           ->_addLimitForQueryDependent($query, $map);

    return $query;
  }

  protected function _addScopeForQueryDependent(Storm_Query $query, array $map): self
  {
    foreach (($map['scope'] ?? []) as $scope_field => $scope_value)
      ($scope_value instanceof Storm_Query_Clause)
        ? $query->addClause($scope_value)
        : $query->eq(strtolower($scope_field), $scope_value);

    return $this;
  }

  protected function _addOrderForQueryDependent(Storm_Query $query, array $map): self
  {
    Storm_Query_Order::parseOrderClause($map['order'] ?? [])
      ->addOrdersInQuery($query);

    return $this;
  }

  protected function _addLimitForQueryDependent(Storm_Query $query, array $map): self
  {
    if (null !== ($limit = ($map['limit'] ?? null)))
      $query->limit((int) $limit);

    return $this;
  }

  protected function _wrapDependentsWithDefinedClass(string $field,
                                                     mixed $dependents): mixed
  {
    if ( ! isset($this->_has_many[$field]['instance_of']))
      return $dependents;

    $support_class = $this->_has_many[$field]['instance_of'];
    if (is_a($dependents, $support_class))
      return $dependents;

    return new $support_class($dependents);
  }

  protected function _setDependents(string $field, mixed $value): self
  {
    if ( ! $value)
      $value = [];

    $map = $this->descriptionOf($field);
    if (isset($map['through'])) {
      $through_field = $map['through'];

      if (isset($this->_belongs_to[$through_field]))
        return ($dependent = $this->_getDependent($through_field))
          ? $dependent->callSetterByAttributeName($field, $value)
          : $this;

      $this->_setDependents($through_field, []);

      foreach ($value as $item)
        $this->_addDependent($field, $item);

      return $this;
    }

    $this->_has_many_attributes[$field] = $this->_wrapDependentsWithDefinedClass($field, $value);
    foreach ($value as $item)
      $item->_set($map['role'], $this);

    return $this;
  }

  protected function _getDependentIdForField(string $field): mixed
  {
    $field_id = $this->getIdFieldForDependent($field);

    if (array_key_exists($field_id, $this->_attributes))
      return $this->_attributes[$field_id];

    if (array_key_exists($field_id, $this->_default_attribute_values))
      return $this->_default_attribute_values[$field_id];

    return null;
  }

  public function callGetterByAttributeName(string $attribute): mixed
  {
    return call_user_func([$this,
                           'get' . $this->attributeNameToAccessor($attribute)]);
  }

  public function callSetterByAttributeName(string $attribute, mixed $value): self
  {
    return call_user_func([$this, 'set' . $this->attributeNameToAccessor($attribute)], $value);
  }

  public function callAdderByAttributeName(string $attribute, $value) {
    return call_user_func([$this, 'add' . $this->attributeNameToAccessor($this->_singularize($attribute))], $value);
  }

  protected function _getDependent(string $field,
                                   bool $find_in_db = true): ?Storm_Model_Abstract
  {
    $through_dependency = array_key_exists('through', $this->_belongs_to[$field]);

    // delegate to a dependent
    if ($through_dependency) {
      $through_field = $this->_belongs_to[$field]['through'];

      return ($dependent = $this->_getDependent($through_field, $find_in_db))
        ? $dependent->callGetterByAttributeName($field)
        : null;
    }

    $id = $this->_getDependentIdForField($field);

    // if the instance is in cache, returns it
    if (array_key_exists($field, $this->_belongs_to_attributes))
      return (($dependent = $this->_belongs_to_attributes[$field])
              && ($id == $dependent->getId()))
        ? $dependent
        : null;

    if (null === $id)
      return null;

    // otherwise in database ?
    return $find_in_db
      ? ($this->_belongs_to_attributes[$field] = $this
         ->_getLoaderForModel($this->_belongs_to[$field]['model'])
         ->find($id))
      : ($this
         ->_getLoaderForModel($this->_belongs_to[$field]['model'])
         ->getLoadedInstances()[$id] ?? null);
  }

  protected function _singularize(string $str): string
  {
    return Storm_Inflector::singularize($str);
  }

  protected function _pluralize(string $str): string
  {
    return Storm_Inflector::pluralize($str);
  }

  protected function _set(string $field, mixed $value): self
  {
    if (array_key_exists($field, $this->_has_many))
      return $this->_setDependents($field, $value);

    if (array_key_exists($field, $this->_belongs_to)) {
      $id_field = $this->getIdFieldForDependent($field);
      $this->_attributes[$id_field] = (null == $value) ? null : $value->getId();
      $this->_belongs_to_attributes[$field] = $value;

      return $this;
    }

    return $this->_setAttributes($field, $value);
  }

  protected function _setAttributes(string $field, mixed $value): self
  {
    $this->_attributes[$field] = $value;

    return $this;
  }

  public function __toString()
  {
    return get_class($this) . '(id=' . $this->getId() . ')';
  }

  public function acceptHierarchyVisitor($visitor): self
  {
    $my_name = get_class($this);

    foreach ([$this->_belongs_to, $this->_has_many] as $links)
      array_walk($links,
                 function ($link, $link_name) use ($my_name, $visitor): void {
                   if ( ! isset($link['model']))
                     return;

                   $visitor->visit(self::getLoaderFor($link['model'])->newInstance(),
                                   $my_name, $link_name);
                 });

    return $this;
  }

  public function dotGraph()
  {
    $visitor = new Storm_Model_HierarchyVisitor;
    $this->acceptHierarchyVisitor($visitor);

    return $visitor->asDot();
  }

  public function isFixedId(): bool
  {
    return $this->_fixed_id;
  }

  public function copyState(?Storm_Model_Abstract $old_instance): self
  {
    if ($old_instance)
      $this->_model_state = $old_instance->_model_state;

    return $this;
  }

  public function hasBeenSaved(): bool
  {
    return $this->_modelState()->hasBeenSaved();
  }

  public function hasBeenUpdated(): bool
  {
    return $this->_modelState()->hasBeenUpdated();
  }

  public function isFixture(): bool
  {
    return $this->_modelState()->isFixture();
  }

  public function existInDatabase(): bool
  {
    return $this->_modelState()->existInDatabase();
  }

  public function hasBeenLoaded(): bool
  {
    return $this->_modelState()->hasBeenLoaded();
  }

  public function hasBeenDeleted(): bool
  {
    return $this->_modelState()->hasBeenDeleted();
  }

  protected function _modelState(): Storm_Model_State
  {
    return $this->_model_state ??= (static::getLoader()->isVolatile()
                                    ? new Storm_Model_State_Volatile
                                    : new Storm_Model_State);
  }

  public function beSaved(): self
  {
    $this->_modelState()->setSaved(true);

    return $this;
  }

  public function beUpdated(): self
  {
    $this->_modelState()->setUpdated(true);

    return $this;
  }

  public function beFixture(): self
  {
    $this->_modelState()->beFixture();

    return $this;
  }

  public function beDeleted(): self
  {
    $this->_modelState()->setDeleted(true);

    return $this;
  }

  public function beLoaded(): self
  {
    $this->_modelState()
         ->resetState()
         ->setLoaded(true);

    return $this;
  }
}
