<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2022 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

*/


class Storm_Model_BaseLoader
{

  protected static $_loader_cache;
  protected static $_default_strategy = 'Db';
  protected static array $_cached_instances = [];
  protected static array $_rand_callback_for_model = [];
  protected string $_model;
  protected array $_loaded_instances = [];
  protected string $_instance_class = '';
  protected ?Storm_Model_PersistenceStrategy_Abstract $_persistence_strategy = null;

  public function __construct(string $class)
  {
    $this->_model = $class;
  }

  public function isBaseLoader(): bool
  {
    return true;
  }

  public function randCallbackForModel(Storm_Query_Instance $a,
                                       Storm_Query_Instance $b): int
  {
    return ($callback = (static::$_rand_callback_for_model[$this->_model] ?? null))
      ? $callback($a, $b)
      : (rand(0, 2) -1);
  }

  public function newFromRow(array $row): Storm_Model_Interface
  {
    $row = array_change_key_case($row, CASE_LOWER);

    return $this
      ->newInstance()
      ->initializeAttributes($row);
  }

  public function newFromRowWithCache(array $row, ?Storm_Model_Abstract $old_instance = null): Storm_Model_Interface
  {
    $instance = $this->newFromRow($row);

    $this->cacheInstance($instance->copyState($old_instance)->beLoaded());

    return $instance;
  }

  public function newInstance(?array $attributes = null): Storm_Model_Interface
  {
    $class = $this->_instance_class ?: $this->_model;
    $instance = new $class;
    $this->_instance_class = '';

    if (null !== $attributes && [] !== $attributes)
      $instance->updateAttributes($attributes);

    return $instance;
  }

  public static function setDefaultStrategy(string $strategy): void
  {
    static::$_default_strategy = $strategy;
  }

  public static function defaultToVolatile(): void
  {
    static::setDefaultStrategy('Volatile');
  }

  public static function defaultToDb(): void
  {
    static::setDefaultStrategy('Db');
  }

  public function setInstanceClass(string $instance_class): self
  {
    $this->_instance_class = $instance_class;

    return $this;
  }

  /**
   * @param $tbl :  Storm_Model_Table | Storm_Test_ObjectWrapper
   */
  public function setTable($tbl): self
  {
    $this->getPersistenceStrategy()->setTable($tbl);

    return $this;
  }

  /**
   * @return Storm_Model_Table | Storm_Test_ObjectWrapper
   */
  public function getTable()
  {
    return $this->getPersistenceStrategy()->getTable();
  }

  public function getModel(): string
  {
    return $this->_model;
  }

  public function getPersistenceStrategy(): Storm_Model_PersistenceStrategy_Abstract
  {
    if ($this->_persistence_strategy !== null)
      return $this->_persistence_strategy;

    $class_name = 'Storm_Model_PersistenceStrategy_' . static::$_default_strategy;
    return $this->_persistence_strategy = new $class_name($this);
  }

  /**
   * Add a model into runtime cache
   */
  public function cacheInstance(Storm_Model_Interface $instance): self
  {
    return $this;
  }

  public function getLoaderCache()
  {
    if ( ! static::$_loader_cache)
      static::$_loader_cache = new Storm_Model_Loader_NullCache;

    return static::$_loader_cache;
  }

  public function beVolatile(): self
  {
    if ( ! $this->getPersistenceStrategy()->isVolatile())
      $this->_persistence_strategy = new Storm_Model_PersistenceStrategy_Volatile($this);

    return $this;
  }

  public function isVolatile(): bool
  {
    return $this->getPersistenceStrategy()->isVolatile();
  }

  public function removeFromCache(array $ids): self
  {
    foreach ($ids as $id)
      if ($model = ($this->_loaded_instances[$id] ?? null))
        $model->beDeleted();

    $this->clearCache($ids);

    return $this;
  }

  public function removeFromCachedInstance(string $key): self
  {
    if ($key)
      unset(static::$_cached_instances[$this->_model][$key]);

    return $this;
  }

  public function fetchForQuery(Storm_Query_Abstract $query,
                                bool $do_cache = false): array
  {
    return ( ! $do_cache
             || $query->notLoadInCache())
      ? $this->_fetchForQuery($query, $do_cache)
      : $this->_fetchForQueryWithoutSelect($query, $do_cache);
  }

  protected function _fetchForQueryWithoutSelect(Storm_Query_Abstract $query,
                                                 bool $do_cache): array
  {
    $result = array_filter((new Storm_Model_PersistenceStrategy_LoadInCache($this))
                           ->fetchForQuery($query));

    return ($result && ($ids = array_unique(array_map(fn($model) => $model->getId(),
                                                      $result)))
            ? [...$result,
               ...$this->_fetchForQuery($query->cloneQuery()
                                        ->not_in($this->getIdField(), $ids), true)]
            : $this->_fetchForQuery($query, true));
  }

  public function countForQuery(Storm_Query_Abstract $query): int
  {
    return $this->_notifyQuery($query)
                ->getPersistenceStrategy()
                ->countForQuery($query);
  }

  public function fetchFirstForQuery(Storm_Query_Abstract $query,
                                     bool $do_cache): mixed
  {
    $result = $do_cache
      ? (new Storm_Model_PersistenceStrategy_LoadInCache($this))
      ->fetchForQuery($query)
      : [];

    if ( ! $result)
      $result = $this->_fetchForQuery($query);

    return $this->_findFirstFrom($result);
  }

  protected function _fetchForQuery(Storm_Query_Abstract $query,
                                    bool $do_cache = false): array
  {
    $cache_key = $do_cache ? $query->uniqKey() : '';

    if (null !== ($instances = $this->_getInstancesFromCachedInstances($query,
                                                                       $cache_key)))
      return $instances;

    return $this->_addToCachedInstances($query, $cache_key, $this->_notifyQuery($query)
                                        ->getPersistenceStrategy()
                                        ->fetchForQuery($query));
  }

  protected function _getInstancesFromCachedInstances(Storm_Query_Abstract $query,
                                                      string $cache_key): ?array
  {
    return $cache_key
      ? (static::$_cached_instances[$this->_model][$cache_key] ?? null)
      : null;
  }

  protected function _addToCachedInstances(Storm_Query_Abstract $query,
                                           string $cache_key,
                                           array $instances): array
  {
    if ($cache_key)
      static::$_cached_instances[$this->_model][$cache_key] = $instances;

    return $instances;
  }

  public function updateForQuery(Storm_Update $update): int
  {
    $this->isVolatile()
      ? $this->_notifyQuery($update)
      : (new Storm_Model_PersistenceStrategy_LoadInCache($this))
      ->updateForQuery($update);

    return $this->getPersistenceStrategy()->updateForQuery($update);
  }

  public function deleteForQuery(Storm_Delete $delete): int
  {
    if ($delete->isBasic()
        || ! ($class_ids = (new $this->_model)->dependentsDeleteClassAndId()))
      return $this->_basicDelete($delete);

    $id_field = $this->getIdField();
    $ids = [];
    $page = 1;
    $query = $delete
      ->promoteToQuery()
      ->select($id_field)
      ->beBasic();

    while ($chunked_ids = array_map(fn($row) => $row[$id_field],
                                    $query->limit_page($page, 1000)
                                    ->fetchAll())) {
      $this->_cascadeDelete($chunked_ids)
           ->_dependentsDeleteAll($class_ids, $chunked_ids);

      $ids = [...$ids, ...$chunked_ids];
      $page++;

      if (count($chunked_ids) < 1000)
        break;
    }

    $count = $ids
      ? $this->_basicDelete($delete)
      : 0;

    return $count;
  }

  protected function _dependentsDeleteAll(array $class_ids, array $ids): self
  {
    if ( ! $class_ids || ! $ids)
      return $this;

    foreach ($class_ids as $class_name => $id_field) {
      $delete_depend_query = $class_name::deleteQuery();

      count($ids) > 1
        ? $delete_depend_query->in($id_field, $ids)
        : $delete_depend_query->eq($id_field, reset($ids));

      $delete_depend_query->deleteAll();
    }

    return $this;
  }

  protected function _basicDelete(Storm_Delete $delete): int
  {
    $this->isVolatile()
      ? $this->_notifyQuery($delete)
      : (new Storm_Model_PersistenceStrategy_LoadInCache($this))
      ->deleteForQuery($delete);

    return $this->getPersistenceStrategy()
                ->deleteForQuery($delete);
  }

  protected function _cascadeDelete(array $ids): self
  {
    return $this;
  }

  protected function _notifyQuery(Storm_Query_Abstract $query): self
  {
    if ($this->isVolatile())
      $query->notify();

    return $this;
  }

  protected function _findFirstFrom(array $instances): mixed
  {
    if ([] === $instances)
      return null;

    $first_instance = reset($instances);
    if ($first_instance instanceof Storm_Model_Abstract)
      $this->cacheInstance($first_instance);

    return $first_instance;
  }
}
