<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2011 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

*/


/**
 * Load data from database for a given model
 *
 * Model has to
 *  - extend Storm_Model_Abstract
 *  - define its table name
 *  - implement static getLoader (will be deprecated when PHP 5.3 compatibility arise)
 *
 * @example
 * class MyModel extends Storm_Model_Abstract {
 *  public $_table_name='my_model';
 *
 *  public static function getLoader() {
 *    return self::getLoaderFor(__CLASS__);
 *  }
 * }
 */


class Storm_Model_Loader extends Storm_Model_BaseLoader
{

  protected string $_table;
  protected array $_loaded_instances = [];
  protected string $_id_field;

  /**
   * @param $cache Zend_Cache
   */
  public static function setDefaultCache($cache): void
  {
    static::$_loader_cache = $cache;
  }

  public static function resetCache(): void
  {
    static::$_cached_instances = [];
    static::$_loader_cache = null;
    static::$_rand_callback_for_model = [];
  }

  public function __construct(string $class)
  {
    parent::__construct($class);

    if (null === ($this->_id_field = strtolower($class::getClassVar('_table_primary'))))
      $this->_id_field = 'id';
  }

  public function isBaseLoader(): bool
  {
    return false;
  }

  public function getIdField(): string
  {
    return $this->_id_field;
  }

  public function randCallback(Closure $callback): void
  {
    static::$_rand_callback_for_model [$this->_model] = $callback;
  }

  public function findAll(mixed $select = null): array
  {
    return $this->getPersistenceStrategy()->findAll($select);
  }

  public function find(mixed $id): ?Storm_Model_Abstract
  {
    if (null === $id)
      return null;

    if ($instance = $this->_findInCache($id))
      return $instance;

    if ($instance = $this->getLoaderCache()->load($this->_model, $id))
      return $instance;

    if ($instance = $this->getPersistenceStrategy()->find($id))
      return $instance->beLoaded();

    $this->_loaded_instances[$id] = null;

    return null;
  }

  protected function _findInCache(mixed $id): ?Storm_Model_Abstract
  {
    if ( ! ($instance = $this->instanceById($id, true)))
      return null;

    return ( ! $this->isVolatile() || $instance->hasBeenLoaded())
      ? $instance
      : null;
  }

  public function getLoadedInstances(bool $filtered = true): array
  {
    $instances = array_filter($this->_loaded_instances);

    return ($this->isVolatile() && $filtered)
      ? $this->_loadedInstancesVolatile($instances)
      : $instances;
  }

  protected function _loadedInstancesVolatile(array $loaded_instances): array
  {
    $instances = [];

    foreach ($loaded_instances as $id => $instance)
      if ( ! $instance->hasBeenDeleted())
        $instances [$id] = $instance;

    return $instances;
  }

  public function clearCache(array $deleted_ids = []): self
  {
    unset(static::$_cached_instances[$this->_model]);

    $instances = [];

    foreach (array_filter($this->_loaded_instances) as $instance)
      if (($cleared_instance = $this->_clearCacheInstance($instance, $deleted_ids))
          && ! $cleared_instance->hasBeenDeleted())
        $instances [$cleared_instance->getId()] = $cleared_instance;

    $this->_loaded_instances = (($deleted_ids || $this->isVolatile())
                                ? $instances
                                : []);

    return $this;
  }

  protected function _clearCacheInstance(Storm_Model_Abstract $instance,
                                         array $deleted_ids): ?Storm_Model_Abstract
  {
    if ($deleted_ids)
      return in_array($instance->getId(), $deleted_ids, true)
        ? $instance->resetCacheAllDependents()
        : $instance;

    return ($this->isVolatile() && $instance->existInDatabase())
      ? $instance->resetCacheAllDependents()
      : null;
  }

  public function instanceById(mixed $id,
                               bool $be_loaded = false): ?Storm_Model_Abstract
  {
    $instance = (null === $id)
      ? null
      : ($this->getLoadedInstances()[$id] ?? null);

    return ($instance && $be_loaded)
      ? $instance->beLoaded()
      : $instance;
  }

  /**
   * Create a new instance and cache it
   */
  public function newInstanceWithId(mixed $id,
                                    ?array $attributes = null): Storm_Model_Abstract
  {
    $instance = $this
      ->newInstance()
      ->setId($id);

    // id must be set BEFORE other attributes (they may depend on the id's value)
    if ( ! empty($attributes))
      $instance->updateAttributes($attributes);

    return $instance;
  }

  public function newInstanceWithIdAssertSave(mixed $id, ?array $attributes = null): Storm_Model_Abstract
  {
    $instance = $this->newInstanceWithId($id, $attributes);
    $instance->assertSave(true);

    return $instance;
  }

  /**
   * Create an instance and assigns its
   * attributes using an associative array
   * (name_of_attribute => value)
   */
  public function newFromRow(array $row): Storm_Model_Interface
  {
    $row = array_change_key_case($row, CASE_LOWER);
    $id = $row[$this->getIdField()];
    unset($row[$this->getIdField()]);

    if ($instance = $this->instanceById($id))
      return $instance;

    return $this
      ->newInstanceWithId($id)
      ->initializeAttributes($row);
  }

  public function cacheInstance(Storm_Model_Interface $instance): self
  {
    $this->_loaded_instances [$instance->getId()] = $instance;

    $this->getLoaderCache()->save($instance);

    return $this;
  }

  /**
   * Insert (if new) or update the record in DB
   *
   * @return int
   */
  public function save(Storm_Model_Abstract $model): int
  {
    $is_changed_id = $this->_isChangedId($model);

    return (new Storm_Model_Loader_SaveOperation($this))->save($model, $is_changed_id);
  }

  protected function _isChangedId(Storm_Model_Abstract $model): bool
  {
    if (null === ($old_id = $model->changedId()))
      return false;

    if ($erase_model = $this->find($model->getId()))
      $erase_model->beDeleted()->forceDelete();

    $this->basicDeleteBy(['id' => $old_id]);
    $this->removeFromCache([$old_id]);

    return true;
  }

  /**
   * Save model and force primary key to be written / skip autoincrement
   */
  public function saveForcePrimaryKey(Storm_Model_Abstract $model): int
  {
    return (new Storm_Model_Loader_SaveForcePrimaryKeyOperation($this))->save($model);
  }

  public function isValueCanBeSaved(mixed $value): bool
  {
    return (new Storm_Model_Loader_SaveOperation($this))->isValueCanBeSaved($value);
  }

  public function delete(Storm_Model_Abstract $model): bool
  {
    $deleted_count = $this->getPersistenceStrategy()
                          ->delete($model);

    if ($deleted_count)
      $this->_cascadeDelete([$model->getId()])
           ->clearCache([$model->beDeleted()->getId()]);

    return 0 < $deleted_count;
  }

  public function deleteBy(array $clauses, int $page_size = 100): int
  {
    $done = 0;
    $ids = [];

    do {
      $clauses['limit'] = $page_size;
      $models = $this->getPersistenceStrategy()->findAllBy($clauses);

      foreach ($models as $model) {
        $ids [] = $model->getId();
        $model->delete();
      }

      $done += count($models);
    } while (count($models) > 0);

    $this->removeFromCache($ids);

    return $done;
  }

  /**
   * This delete operation totally ignore hooks and cascading.
   * But this is faster. Use with caution
   */
  public function basicDeleteBy(mixed $clauses): int
  {
    $count = $this->getPersistenceStrategy()->deleteBy($clauses);

    $this->clearCache();

    return $count;
  }

  public function truncate(): bool
  {
    return $this->getPersistenceStrategy()->truncate();
  }

  public function getIdFieldForDependent(string $field): string
  {
    $model_instance = new $this->_model;

    return $model_instance->getIdFieldForDependent($field);
  }

  /**
   * @example
   * NewsletterSubscription::getLoader()->findAllBy([
   *    'role' => 'newsletter',
   *    'model' => $show
   * ]);
   * Return related rows
   *
   * @example
   * Book::getLoader()->findAllBy(array(
   *    'order' => 'creation_date desc',
   *    'limit' => 10,
   *    'where' => 'note>2',
   *    'title like' => '%apple%',
   *    'left(isbn, 3) => '978',
   *    'tag' => null,
   *    'limitPage' => array($page_number, $items_by_page)
   * ));
   * Return first 10 rows ordered by creation_date where note > 2 and tag is null
   *
   * @example
   * BlueRay::getLoader()->findAllBy(array(
   *    'alpha_key' => 'MILLENIUM',
   *    'user_id' => '12'
   * ));
   * Return rows where alpha_key = 'MILLENIUM' and user_id = '12'
   */
  public function findAllBy(array $args): array
  {
    $filter_given_func = $args['callback'] ?? null;

    $filter_func = null;
    if (isset($args['select'])) {
      $select = $args['select'];
      $filter_func = fn($model) => $model->$select();
    }

    unset($args['select']);
    unset($args['callback']);

    $filter = $filter_func ?: $filter_given_func;

    $result = $this->getPersistenceStrategy()->findAllBy($args);
    return isset($filter)
      ? array_filter($result, $filter)
      : $result;
  }

  /**
   * Execute an update query on all rows matching $where
   * and apply new $values
   * @param array $where update where condition
   * @param array $set_values update set statement
   * @return int number of affected rows
   * @example
   * Video::updateAllBy(['support' => 'vhs',
   *                     'where' => 'year < 2000'],
   *                    ['support' => 'dvd']);
   */
  public function updateAll(array $where, array $set_values): int
  {
    $this->clearCache();

    return $this->getPersistenceStrategy()->updateAll($where, $set_values);
  }

  /**
   * Sends a count request and returns the value
   * @see findAllBy
   */
  public function countBy(array $args): int
  {
    return $this->getPersistenceStrategy()->countBy($args);
  }

  /**
   * return number of this model in db
   * @see countBy
   */
  public function count(): int
  {
    return $this->countBy([]);
  }

  public function findFirstBy(array $args): ?Storm_Model_Abstract
  {
    $instances = $this->findAllBy($this->_prepareFirstBy($args));

    return $this->_findFirstFrom($instances);
  }

  protected function _prepareFirstBy(array $args): array
  {
    $args['limit'] = 1;

    return $args;
  }

  public function fetchAllBy(array $fields, array $params = []): array
  {
    return $this->getPersistenceStrategy()->fetchAllBy($fields, $params);
  }

  public function query(?string $alias = null): Storm_Query
  {
    return new Storm_Query($this, $alias);
  }

  public function orQuery(?string $alias = null): Storm_Query
  {
    return $this->query()->beOr();
  }

  public function join(string $alias): Storm_Join
  {
    return new Storm_Join($this, $alias);
  }

  public function orJoin(string $alias): Storm_Join
  {
    return $this->join()->beOr();
  }

  public function inner(string $alias): Storm_Join_Inner
  {
    return new Storm_Join_Inner($this, $alias);
  }

  public function left(string $alias): Storm_Join_Left
  {
    return new Storm_Join_Left($this, $alias);
  }

  public function right(string $alias): Storm_Join_Right
  {
    return new Storm_Join_Right($this, $alias);
  }

  public function subIn(?string $alias = null): Storm_Query_In
  {
    return new Storm_Query_In($this, $alias);
  }

  public function subExists(?string $alias = null): Storm_Query_Exists
  {
    return new Storm_Query_Exists($this, $alias);
  }

  public function updateQuery(): Storm_Update
  {
    return new Storm_Update($this, null);
  }

  public function deleteQuery(): Storm_Delete
  {
    return new Storm_Delete($this, null);
  }
}
