<?php
/*
STORM is under the MIT License (MIT)

Copyright (c) 2010-2011 Agence Française Informatique http://www.afi-sa.fr

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/


class Storm_Model_PersistenceStrategy_Db
  extends Storm_Model_PersistenceStrategy_Abstract
{

  public function find(mixed $id): ?Storm_Model_Abstract
  {
    $rowset = $this->getTable()->find($id)->toArray();

    return (count($rowset) > 0)
      ? $this->_loader->newFromRowWithCache(reset($rowset))
      : null;
  }

  public function update(array $data, mixed $id): int
  {
    return $this->_updateSql($data, $this->_loader->getIdField() . "='" . $id . "'");
  }

  public function updateAll(array $where, array $set_values): int
  {
    return $this->_updateSql($set_values, $this->_generateWhereClause($where));
  }

  protected function _updateSql(array $set_values, string $where): int
  {
    return $this->getTable()->update($set_values, $where);
  }

  public function delete(Storm_Model_Abstract $model): int
  {
    return $this->deleteBy([$this->_loader->getIdField() => $model->getId()]);
  }

  public function deleteBy(mixed $where_or_clauses): int
  {
    return is_array($where_or_clauses)
      ? $this->getTable()->delete($this->_generateWhereClause($where_or_clauses))
      : $this->getTable()->delete($where_or_clauses);
  }

  protected function _generateWhereClause(array $where_or_clauses): string
  {
    $where = [];

    foreach ($where_or_clauses as $key => $value)
      $where [] = $this->_generateWhereClauseForKeyAndValue($key, $value);

    return implode(' and ', $where);
  }

  protected function _generateWhereClauseForKeyAndValue(string $key,
                                                        mixed $value): string
  {
    $clause = ($value instanceof Storm_Query_Clause)
      ? $value
      : Storm_Query_Clause::newWith($key, $value);

    return $clause->getFormatDb();
  }

  /** @see Zend_Db_Adapter_Abstract::quoteInto() */
  public function quoteInto($text, $value, $type = null, $count = null) {
    return $this->getTable()->getAdapter()->quoteInto($text, $value, $type, $count);
  }

  public function fetchForQuery(Storm_Query_Abstract $query): array
  {
    $select = $this->_assembleQuery($query);

    if ($query->isBasic())
      return ($rowset = $this->getTable()->fetchAll($select))
        ? $rowset->toArray()
        : [];

    return $this->findAll($select);
  }

  public function countForQuery(Storm_Query_Abstract $query): int
  {
    $rows = $this->fetchForQuery($query->beBasic());

    return $rows[0]['numberof'] ?? 0;
  }

  public function updateForQuery(Storm_Update $update): int
  {
    return Zend_Db_Table_Abstract::getDefaultAdapter()->update($update->assembleFrom(),
                                                               $update->assembleSets(),
                                                               $update->assembleWhere());
  }

  public function deleteForQuery(Storm_Delete $delete): int
  {
    return Zend_Db_Table_Abstract::getDefaultAdapter()->delete($delete->assembleFrom(),
                                                               $delete->assembleWhere());
  }

  protected function _assembleQuery(Storm_Query_Abstract $query): Zend_Db_Table_Select
  {
    $select = $query->getAssembleSelect();
    $this->setTable($select->getTable());

    return $select;
  }

  public function findAll(mixed $select = null): array
  {
    if (is_string($select))
      return $this->_findAllByString($select);

    return ($rowset = $this->getTable()->fetchAll($select))
      ? array_map(fn($row) => $this->_loader->newFromRowWithCache($row),
                  $rowset->toArray())
      : [];
  }

  protected function _findAllByString(mixed $query): array
  {
    return array_map(fn($row) => $this->_loader->newFromRowWithCache($row),
                     $this->getTable()->getAdapter()->fetchAll($query));
  }

  public function findAllBy(array $args = []): array
  {
    return ($select = $this->_generateSelectFor($args))
      ? $this->findAll($select)
      : [];
  }

  public function fetchAllBy(array $fields, array $params = []): array
  {
    if ( ! ($select = $this->_generateSelectFor($params)))
      return [];

    $select->from($select->getTable(), $fields);
    return ($rowset = $this->getTable()->fetchAll($select))
      ? $rowset->toArray()
      : [];
  }

  public function countBy(array $args = []): int
  {
    return (($select = $this->_generateSelectFor($args))
            ? $this->_count($select)
            : 0);
  }

  protected function _count(mixed $select): int
  {
    $select->from($this->getTable(),
                  ['count(*) as numberof']);

    $rows = $this->getTable()->fetchAll($select)->toArray();

    return $rows[0]['numberof'];
  }

  public function insert(array $data): bool
  {
    return (bool) $this->getTable()->insert($data);
  }

  public function lastInsertId(): mixed
  {
    return $this->getTable()->getAdapter()->lastInsertId();
  }

  /**
   * @see findAllBy
   */
  public function _generateSelectFor(array $args = []): mixed
  {
    $select = $this->getTable()->select();

    foreach ($args as $field => $value)
      $this->_addInSelect($select, $field, $value);

    return $select;
  }

  protected function _addInSelect(mixed $select,
                                  string $field,
                                  mixed $value_or_clause): self
  {
    if (in_array($field, ['order', 'limit', 'where'], true)) {
      $select->$field($value_or_clause);

      return $this;
    }

    if (in_array($field, [Storm_Query_Clause::CLAUSE_LIMIT_PAGE, 'scope'], true))
      return $this->$field($select, $value_or_clause);

    if ('group_by' === $field) {
      $select->group($value_or_clause);

      return $this;
    }

    $select->where($this->_generateWhereClauseForKeyAndValue($field,
                                                             $value_or_clause));
    return $this;
  }

  protected function limitPage(mixed $select, mixed $value): self
  {
    $clause = Storm_Query_Clause::limitPage($value);

    $clause->assemble($select);

    return $this;
  }

  protected function scope(mixed $select, array $value): self
  {
    foreach ($value as $scope_field => $scope_value)
      $select->where($scope_field . '=?', $scope_value);

    return $this;
  }

  public function truncate(): bool
  {
    $table = $this->getTable();

    return $table->getAdapter()->truncate($table->getName());
  }
}
