<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2011 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

*/


class Storm_Model_PersistenceStrategy_Volatile
  extends Storm_Model_PersistenceStrategy_Abstract
{
  use Storm_Model_PersistenceStrategy_QueryStrategy;

  public function isVolatile(): bool
  {
    return true;
  }

  public function findAll(mixed $select = null): array
  {
    $select = $this->_extractRoleAndModel($select);

    return null === $select ? [] : $this->_findAll($select);
  }

  protected function _findAll(mixed $select): array
  {
    $group_by = $this->_getGroupBy($select);
    $order = $this->_getOrder($select);
    $limit = trim($this->_getLimit($select));
    $limit_page = $this->_getLimitPage($select);

    $values = [];
    $this->_allMatchingInstancesDo($this->_modelsFromClauses($select),
                                   function($model) use (&$values): void {
                                     $values [] = $model;
                                   });

    $values = $this->_groupBy($values, $group_by);
    $values = $this->_ordered($values, $order);
    $values = $this->_limited($values, $limit);

    if ($limit_page)
      $values = array_slice($values, $limit_page[1], $limit_page[0]);

    return $this->_instancesFor($values);
  }

  protected function _instancesFor(array $values): array
  {
    if ( ! $values)
      return [];

    if ($this->_loader->isBaseLoader())
      return array_map(fn($row) => $this->_loader->newFromRowWithCache($row),
                       $values);

    $id_field = $this->_loader->getIdField();

    return array_map(fn($row) => $this->_newFromRow($row, $id_field),
                     $values);
  }

  protected function _newFromRow(array $row, string $id_field): Storm_Model_Interface
  {
    $id = $data[$id_field] ?? null;

    $instance = (null === $id)
      ? null
      : $this->_loader->instanceById($id);

    return ($instance && $instance->hasBeenLoaded())
      ? $instance
      : $this->_loader->newFromRowWithCache($row, $instance);
  }

  public function fetchAllBy(array $fields, array $params = []): array
  {
    return array_map(function($model) use($fields)
    {
      $keys = array_map(fn($field) => current(explode(' as ', $field)),
                        $fields);

      $aliases = array_map(function($field)
      {
        $parts = explode(' as ', $field);
        return $parts[1] ?? current($parts);
      },
                           $fields);

      $map = array_combine($keys, $aliases);

      $filtered = array_filter($model->getRawAttributes(),
                               fn($key) => in_array($key, $keys),
                               ARRAY_FILTER_USE_KEY);

      $row = [];
      foreach($filtered as $k => $v)
        $row[$map[$k]] = $v;

      return $row;
    },
                     $this->findAll($params));
  }

  protected function _extractRoleAndModel(mixed $select): array
  {
    $select = $this->_strtolowerkeys($select);

    if (array_key_exists('scope', $select)) {
      $select = array_merge($select,
                            array_change_key_case($select['scope']));
      unset($select['scope']);
    }

    return $select;
  }

  protected function _strtolowerkeys(mixed $select): array
  {
    if ( ! is_array($select))
      return [];

    if ( ! isset($select[Storm_Query_Clause::CLAUSE_LIMIT_PAGE]))
      return array_change_key_case($select);

    $lower_select = [];

    foreach ($select as $key => $value)
      $lower_select [Storm_Query_Clause::CLAUSE_LIMIT_PAGE == $key
                     ? $key : strtolower($key)] = $value;

    return $lower_select;
  }

  protected function _getInstancesArray(): array
  {
    return
      array_values(
                   array_map(fn($model) => $model->getRawAttributes(),
                             $this->_loader->getLoadedInstances()));
  }

  public function findAllBy(array $args = []): array
  {
    return $this->findAll($args);
  }

  protected function _ordered(array $result, mixed $order): array
  {
    if ( ! $order || ! $result)
      return $result;

    $query_order = Storm_Query_Order::parseOrderClause($order);

    usort($result, fn($a, $b) => $query_order->compareVolatile($a, $b));
    return $result;
  }

  protected function _limited(array $result, string $limit): array
  {
    if ('' === $limit)
      return $result;

    if (0 == $limit)
      return [];

    if (false === strpos($limit, ','))
      return array_slice($result, 0, (int)$limit);

    $parts = explode(',', $limit);
    return array_slice($result, (int)$parts[0], (int)$parts[1]);
  }

  protected function _groupBy(array $result, string $group_by): array
  {
    if ($group_by === '' || $group_by === '0')
      return $result;

    $seen_values = [];
    return array_filter($result,
                        function($row) use ($group_by, &$seen_values)
                        {
                          if (in_array($row[$group_by], $seen_values))
                            return false;

                          $seen_values [] = $row[$group_by];
                          return true;
                        });
  }

  public function containsAllAttributes(array $model, array $select): bool
  {
    foreach ($select as $key => $value)
      if ( ! $this->containsAttribute($model, $key, $value))
        return false;

    return true;
  }

  public function containsAttribute(array $model, string $key, mixed $value): bool
  {
    $clause = ($value instanceof Storm_Query_Clause)
      ? $value
      : null;

    if ( ! $clause && preg_match('/left\((.+),(.+)\)/', $key, $matches))
      $clause = Storm_Query_Clause::like(trim($matches[1]),
                                         (substr($value, 0, (int)$matches[2]) . '%'));

    if ( ! $clause)
      $clause = Storm_Query_Clause::newWith($key, $value);

    return $clause->containAttibuteInVolatile($model);
  }

  public function find(mixed $id): ?Storm_Model_Abstract
  {
    return $this->_loader->instanceById($id);
  }

  public function insert(array $data): bool
  {
    if ( ! isset($data[$this->_loader->getIdField()]))
      $data [$this->_loader->getIdField()] = $this->lastInsertId();

    $this->_loader->newFromRow($data);

    return true;
  }

  public function lastInsertId(): int
  {
    $loaded_instances = $this->_loader->getLoadedInstances(false);

    return (count($loaded_instances) < 1)
      ? 1
      : ((int) max(array_keys($loaded_instances)) + 1);
  }

  public function update(array $data, mixed $id): int
  {
    if ( ! ($instance = $this->_loader->instanceById($id)))
      return 0;

    $instance
      ->updateAttributes($data)
      ->setId($id);

    return 1;
  }

  public function delete(Storm_Model_Abstract $model): bool
  {
    return true;
  }

  public function updateAll(array $where, array $set_values): int
  {
    $callback = fn($model) => ($this->_loader->instanceById($model['id'])
                               ->beUpdated()
                               ->initializeAttributes($set_values));

    return $this->_allMatchingInstancesDo($this->_modelsFromClauses($where),
                                          $callback);
  }

  public function deleteBy(array $clauses): int
  {
    $callback = function ($model): ?Storm_Model_Abstract {
      if ($instance = $this->_loader->instanceById($model['id'] ?? 0))
        $instance->beDeleted();

      return $instance;
    };

    return $this->_allMatchingInstancesDo($this->_modelsFromClauses($clauses),
                                          $callback);
  }

  protected function _allMatchingInstancesDo(array $matching_instances,
                                             callable $callback): int
  {
    return count(array_map($callback, $matching_instances));
  }

  protected function _modelsFromClauses(array $clauses): array
  {
    return
      array_filter($this->_getInstancesArray(),
                   fn($model) => $this->containsAllAttributes($model, $clauses));
  }

  public function countBy(array $args): int
  {
    return count($this->findAll($args));
  }

  protected function _getLimit(array &$select): string
  {
    if ( ! isset($select['limit']))
      return '';

    $clause = Storm_Query_Clause::limit($select['limit']);

    unset($select['limit']);
    return $clause->getValue();
  }

  protected function _getLimitPage(array &$select): array
  {
    $clause = Storm_Query_Clause::limitPage($select[Storm_Query_Clause::CLAUSE_LIMIT_PAGE] ?? []);

    unset($select[Storm_Query_Clause::CLAUSE_LIMIT_PAGE]);
    return $clause->getValue();
  }

  protected function _getOrder(array &$select): mixed
  {
    $order = $select['order'] ?? '';
    unset($select['order']);

    return $order;
  }

  protected function _getGroupBy(array &$select): string
  {
    $group_by = $select['group_by'] ?? '';
    unset($select['group_by']);

    return $group_by;
  }

  public function truncate(): bool
  {
    return (bool) $this->deleteBy([]);
  }
}
