<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2022 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

*/


class Storm_Model_Row implements Storm_Model_Interface
{

  const GETTER_METHOD = 'get';

  protected ?Storm_Model_State $_model_state = null;
  protected static array $_loaders = [];
  protected array $_attributes = [];
  protected string $_table_primary = '';
  protected Storm_Model_Associations $_associations;

  public function __construct()
  {
    $this->_associations = new Storm_Model_Associations;
  }

  /**
   * Main purpose is to setup generic getters and setters:
   *
   * $car->getColor();
   *
   * @param string $method
   * @param array $args
   * @return Storm_Model_Abstract
   * @throws Exception
   */
  public function __call($method, $args)
  {
    return $this->_associations->handleCall($this, $method, $args);
  }

  /**
   * Create compatibility with accessor like $miles->instrument.
   * ex:
   *    $miles->setInstrument("trumpet");
   *    assert("trumpet" === $miles->getTrumpet())
   *    assert("trumpet" === $miles->trumpet)
   *
   * If attribute not defined, returns null for compatibility purpose
   * with legacy code - Remember STORM is intended to get
   * easily into messy crappy code. Sorry.
   *
   *     assert(null ===  $miles->zork)
   */
  public function __get(string $field): mixed
  {
    return ($this->_attributes[strtolower($field)]
            ?? $this->_attributes[strtoupper($field)]
            ?? null);
  }

  /**
   * Return the value of attribute $field. If cannot find it, raise Exception.
   * @throws Storm_Model_Exception
   */
  protected function _get(string $field): mixed
  {
    if (array_key_exists($field, $this->_attributes))
      return $this->_attributes[$field];

    throw new Storm_Model_Exception(sprintf('Tried to call unknown method %s::get%s',
                                            get_class($this),
                                            $this->attributeNameToAccessor($field)));
  }

  public static function getClassVar(string $var)
  {
    $reflection = new ReflectionClass(static::class);
    $class_vars = $reflection->getdefaultProperties();

    return $class_vars[$var];
  }

  public static function getLoader()
  {
    return static::getLoaderFor(static::class);
  }

  public static function __callStatic($name, $args)
  {
    return call_user_func_array([static::getLoader(), $name], $args);
  }

  public static function getLoaderFor(string $class)
  {
    return static::$_loaders[$class] ??= new Storm_Model_BaseLoader($class);
  }

  public static function setLoaderFor(string $class, $loader)
  {
    return null;
  }

  public static function unsetLoaders(): void
  {}

  public static function getLoaders(): array
  {
    return static::$_loaders;
  }

  public function getClassName(): string
  {
    return get_class($this);
  }

  /**
   * Return an associative array with attribute
   * name as $key and its value.
   *
   * Used by Loader while saving in order to build the
   * SQL query.
   */
  public function attributesToArray(): array
  {
    return $this->_attributes;
  }

  public function getRawAttributes(): array
  {
    return $this->_attributes;
  }

  /**
   * Return an associative array with attribute
   * name as $key and its value.
   *
   * This method may be redefined in subclasses
   * in order to provide some coupling with Zend_Form::populate()
   */
  public function toArray(): array
  {
    return $this->_attributes;
  }

  public function isAttributeEmpty(string $attribute): bool
  {
    return empty($this->_attributes[$attribute] ?? null);
  }

  /**
   * See __call
   *
   * ex:
   * $car->perform('getColor')
   */
  public function perform(string $method, array $args = []): mixed
  {
    if ( ! $matches = $this->_methodMatch($method))
      throw new Storm_Model_Exception('Tried to call unknown method '
                                      . get_class($this) . '::' . $method);

    return (static::GETTER_METHOD === $matches[1]
            ? $this->_get($this->_accessorToAttributeName($matches[2]))
            : $this);
  }

  protected function _methodMatch(string $method): ?array
  {
    return (static::GETTER_METHOD === substr($method, 0, 3)
            ? [1 => static::GETTER_METHOD,
               2 => substr($method, 3)]
            : null);
  }

  /**
   * UserId -> user_id
   */
  protected function _accessorToAttributeName(string $accessor): string
  {
    return Storm_Inflector::underscorize($accessor);
  }

  /**
   * Set initial attribute values with a [name] => value formatted array
   * Used by Loader::newFromRow
   * The difference with updateAttributes is that it doesn't call
   * setAttributeName(...) magic.
   *
   * So we are sure that attributes are exactly those from db.
   */
  public function initializeAttributes(array $datas): self
  {
    $this->_attributes = array_merge($this->_attributes,
                                     array_change_key_case($datas));

    return $this;
  }

  public function attributeNameToAccessor(string $name): string
  {
    return Storm_Inflector::camelize($name);
  }

  public function isAttributeExists(string $field): bool
  {
    return array_key_exists($field, $this->_attributes);
  }

  public function callGetterByAttributeName(string $attribute): mixed
  {
    return call_user_func([$this, static::GETTER_METHOD
                           . $this->attributeNameToAccessor($attribute)]);
  }

  public function __toString()
  {
    return get_class($this);
  }

  public function copyState(?Storm_Model_Abstract $old_instance): self
  {
    return $this;
  }

  public function hasBeenSaved(): bool
  {
    return false;
  }

  public function hasBeenUpdated(): bool
  {
    return false;
  }

  public function isFixture(): bool
  {
    return false;
  }

  public function existInDatabase(): bool
  {
    return false;
  }

  public function hasBeenDeleted(): bool
  {
    return false;
  }

  public function hasBeenLoaded(): bool
  {
    return true;
  }

  protected function _modelState(): Storm_Model_State
  {
    return $this->_model_state ??= ($this->getLoader()->isVolatile()
                                    ? new Storm_Model_State_Volatile
                                    : new Storm_Model_State);
  }

  public function beSaved(): self
  {
    return $this;
  }

  public function beUpdated(): self
  {
    return $this;
  }

  public function beFixture(): self
  {
    return $this;
  }

  public function beDeleted(): self
  {
    return $this;
  }

  public function beLoaded(): self
  {
    return $this;
  }
}
