<?php
/*
STORM is under the MIT License (MIT)

Copyright (c) 2010-2022 Agence Française Informatique http://www.afi-sa.fr

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/


class Storm_Query_Clause {

  protected Storm_Query_Key $_key;
  protected string $_operator;
  protected string $_identifier = '';
  protected string $_alias = '';
  protected bool $_negated = false;
  protected $_value;

  const
    CLAUSE_WHERE = 'where',
    CLAUSE_LIKE = 'LIKE',
    CLAUSE_EQUAL = '=',
    CLAUSE_NOT_EQUAL = '!=',
    CLAUSE_IN = 'IN',
    CLAUSE_IS = 'IS',
    CLAUSE_GREATER = '>',
    CLAUSE_GREATER_EQUAL = '>=',
    CLAUSE_LESSER = '<',
    CLAUSE_LESSER_EQUAL = '<=',
    CLAUSE_MATCH = 'MATCH',
    CLAUSE_LIMIT = 'limit',
    CLAUSE_LIMIT_PAGE = 'limitPage',
    CLAUSE_ORDER_BY = 'order by',
    CLAUSE_GROUP_BY = 'group by',
    CLAUSE_NOT = ' NOT ',
    CLAUSE_EXISTS = 'EXISTS',
    PERCENT = '%';

  const CLAUSE_IMPLEMENTATIONS =
    [
     self::CLAUSE_LIKE => Storm_Query_ClauseLike::class,
     self::CLAUSE_EQUAL => Storm_Query_ClauseEqual::class,
     self::CLAUSE_IN => Storm_Query_ClauseIn::class,
     self::CLAUSE_IS => Storm_Query_ClauseIsNull::class,
     self::CLAUSE_GREATER => Storm_Query_ClauseGreater::class,
     self::CLAUSE_GREATER_EQUAL => Storm_Query_ClauseGreaterEqual::class,
     self::CLAUSE_LESSER => Storm_Query_ClauseLesser::class,
     self::CLAUSE_LESSER_EQUAL => Storm_Query_ClauseLesserEqual::class,
     self::CLAUSE_MATCH => Storm_Query_Clause_Match::class,
     self::CLAUSE_LIMIT => Storm_Query_ClauseLimit::class,
     self::CLAUSE_LIMIT_PAGE => Storm_Query_ClauseLimitPage::class,
     self::CLAUSE_GROUP_BY => Storm_Query_ClauseGroupBy::class,
     self::CLAUSE_WHERE => Storm_Query_ClauseWhere::class
    ];

  /**
   * @param $value null | string | Storm_Query_Clause
   */
  public function __construct(Storm_Query_Key $key,
                              string $operator,
                              $value) {
    $this->_key = $key;
    $this->_operator = $operator;
    $this->_value = $value;

    $this->_prepareValue();
  }


  protected function _prepareValue() : self {
    return $this;
  }


  public static function newFor(Storm_Query_Key $key,
                                string $operator,
                                $value) : self {
    $class = static::CLAUSE_IMPLEMENTATIONS[$operator];

    return new $class($key, $operator, $value);
  }


  public static function newWith(string $key, $value) : self {
    $operator = static::CLAUSE_EQUAL;
    if (static::CLAUSE_WHERE === $key)
      $operator = static::CLAUSE_WHERE;

    if (' <' === substr($key, -2))
      return static::newFor(new Storm_Query_Key(substr($key, 0, -2)), static::CLAUSE_LESSER, $value);

    if (' >' === substr($key, -2))
      return static::newFor(new Storm_Query_Key(substr($key, 0, -2)), static::CLAUSE_GREATER, $value);

    if (' <=' === substr($key, -3))
      return static::newFor(new Storm_Query_Key(substr($key, 0, -3)), static::CLAUSE_LESSER_EQUAL, $value);

    if (' >=' === substr($key, -3))
      return static::newFor(new Storm_Query_Key(substr($key, 0, -3)), static::CLAUSE_GREATER_EQUAL, $value);

    $is_like = ' like' === substr($key, -5);
    $key = $is_like ? substr($key, 0, strlen($key) - 5) : $key;
    $negated = (' not' === substr($key, -4));
    $key = $negated ? substr($key, 0, strlen($key) - 4) : $key;

    if ($is_like)
      $operator = static::CLAUSE_LIKE;

    $matches = null;
    if (!$is_like)
      preg_match('/^left\(([^,]+),\s*(\d+)\)$/i', $key, $matches);
    if ($matches)
      return (static::newFor(Storm_Query_Key::left(trim($matches[1]), (int)$matches[2]),
                             $operator,
                             $value))
        ->setNegated($negated);

    if (null === $value)
      $operator = static::CLAUSE_IS;

    if (is_array($value))
      $operator = static::CLAUSE_IN;

    return (static::newFor(new Storm_Query_Key($key), $operator, $value))
      ->setNegated($negated);
  }


  public function cloneClause(): self
  {
    return $this;
  }


  public static function isNull(string $key) : self {
    return static::newFor(static::_initKey($key), static::CLAUSE_IS, null);
  }


  public static function in(string $key, array $array) : self {
    return static::newFor(static::_initKey($key), static::CLAUSE_IN, $array);
  }


  public static function equal($key, $value) : self {
    return static::newFor(static::_initKey($key), static::CLAUSE_EQUAL, $value);
  }


  public static function left(string $key, int $left, string $value) : self {
    return static::newFor(static::_initKey($key)
                          ->addLeft($left),
                          static::CLAUSE_EQUAL, $value);
  }


  public static function like(string $key, string $value) : self {
    return static::newFor(static::_initKey($key), static::CLAUSE_LIKE, $value);
  }


  public static function start(string $key, string $value) : self {
    return static::newFor(static::_initKey($key),
                          static::CLAUSE_LIKE,
                          $value . static::PERCENT);
  }


  public static function end(string $key, string $value) : self {
    return static::newFor(static::_initKey($key),
                          static::CLAUSE_LIKE,
                          static::PERCENT . $value);
  }


  public static function greater($key, $value) : self {
    return static::newFor(static::_initKey($key), static::CLAUSE_GREATER, $value);
  }


  public static function greaterEqual($key, $value) : self {
    return static::newFor(static::_initKey($key),
                          static::CLAUSE_GREATER_EQUAL,
                          $value);
  }


  public static function lesser($key, $value) : self {
    return static::newFor(static::_initKey($key), static::CLAUSE_LESSER, $value);
  }


  public static function lesserEqual($key, $value) : self {
    return static::newFor(static::_initKey($key),
                          static::CLAUSE_LESSER_EQUAL,
                          $value);
  }


  public static function match(Storm_Query_MatchRating $match) : self {
    return static::newFor(static::_initKey($match->keyOrArray()),
                          static::CLAUSE_MATCH,
                          $match);
  }


  /**
   * @param $limit int | string
   */
  public static function limit($limit) : self {
    return static::newFor(static::_initKey(static::CLAUSE_LIMIT),
                          static::CLAUSE_LIMIT,
                          $limit);
  }


  /**
   * @param $range array | int | string
   */
  public static function limitPage($range) : self {
    return static::newFor(static::_initKey(static::CLAUSE_LIMIT_PAGE),
                          static::CLAUSE_LIMIT_PAGE,
                          $range);
  }


  public static function group(string $value) : self {
    return static::newFor(static::_initKey(static::CLAUSE_GROUP_BY),
                          static::CLAUSE_GROUP_BY,
                          $value);
  }


  public static function order(string $key) : self {
    return new Storm_Query_Clause_OrderBy(static::_initKey($key), static::CLAUSE_ORDER_BY, null);
  }


  public static function orderMatch($key_or_array, Storm_Query_Clause_Match $match) : self {
    return new Storm_Query_Clause_OrderBy_Match(static::_initKey($key_or_array),
                                                static::CLAUSE_ORDER_BY,
                                                $match);
  }


  public static function orderValue(string $key, $value) : self {
    return new Storm_Query_Clause_OrderBy_Value(static::_initKey($key),
                                                static::CLAUSE_ORDER_BY,
                                                $value);
  }


  public static function orderCast(string $key) : self {
    return new Storm_Query_Clause_OrderBy_Cast(static::_initKey($key),
                                               static::CLAUSE_ORDER_BY,
                                               null);
  }


  public static function orderField(string $key, array $values) : self {
    return new Storm_Query_Clause_OrderBy_Field(static::_initKey($key),
                                                static::CLAUSE_ORDER_BY,
                                                $values);
  }


  protected static function _initKey($key) : Storm_Query_Key {
    return ($key instanceof Storm_Query_Key)
      ? $key
      : new Storm_Query_Key($key);
  }


  public function setNegated(bool $negated) : self {
    $this->_negated = $negated;

    return $this;
  }


  public function getValue() {
    return $this->_value;
  }


  public function withHelperDo(Storm_Query_Helper $helper) : self {
    if ($helper->getIdentifier() !== $this->_identifier)
      $this->_identifier = $helper->getIdentifier();

    if ($helper->getAlias() !== $this->_alias)
      $this->_alias = $helper->getAlias();

    return $this;
  }


  public function computeErrors() : array {
    return [];
  }


  public function filterWith(Storm_Query_Instances $query_instances) : self {
    return $this;
  }


  public function instances(?array $instances = []) : array {
    return $instances;
  }


  public function helperBy(string $identifier) : ?Storm_Query_Helper {
    return null;
  }


  public function assemble(Zend_Db_Table_Select $select) : self {
    return $this;
  }


  public function getFormatDb(?bool $with_quote = true) : string {
    if (static::CLAUSE_WHERE === $this->_operator)
      return '(' . $this->getValue() . ')';

    return $with_quote
      ? Zend_Db_Table_Abstract::getDefaultAdapter()
      ->quoteInto($this->_clauseFormatDb(),
                  $this->getValue(), null, null)
      : str_replace('?', $this->getValue(), $this->_clauseFormatDb());
  }


  public function containAttibuteInVolatile(array $model) : bool {
    if (!$this->_key->existInModel($model))
      return $this->_negated;

    return $this->contains($this->_key->valueForModel($model), $this->getValue());
  }


  public function contains($value_1, $value_2) : bool {
    return ($this->_negated !== ($value_1 == $value_2));
  }


  protected function _clauseFormatDb() : string {
    return $this->_key->key($this->_identifier) . $this->_getOperator() . '?';
  }


  protected function _getOperator() : string {
    return ($this->_negated ? static::CLAUSE_NOT : ' ') . $this->_operator . ' ';
  }
}




class Storm_Query_ClauseLike extends Storm_Query_Clause {

  protected function _prepareValue() : self {
    $this->_value = (string)$this->_value;

    return $this;
  }


  public function contains($value_1, $value_2) : bool {
    $matches = preg_match(('/^' . str_replace('%', '.*', preg_quote($value_2, '/')) . '$/i'),
                          $value_1);

    return $this->_negated ? !$matches : $matches;
  }
}




class Storm_Query_ClauseEqual extends Storm_Query_Clause {

  protected function _getOperator() : string {
    return ' ' . ($this->_negated ? static::CLAUSE_NOT_EQUAL : $this->_operator) . ' ';
  }
}




class Storm_Query_ClauseIn extends Storm_Query_Clause {

  protected function _prepareValue() : self {
    if (!is_array($this->_value))
      $this->_value = [];

    return $this;
  }


  public function containAttibuteInVolatile(array $model) : bool {
    if (!$this->_key->existInModel($model))
      return $this->_negated;

    if (0 === count($this->getValue()))
      throw new Storm_Model_Exception(sprintf('array given for %s is empty',
                                              $this->_key->getKey()));

    return $this->contains($this->_key->valueForModel($model), $this->getValue());
  }


  public function contains($value_1, $value_2) : bool {
    if ( ! is_array($value_2))
      $value_2 = [$value_2];

    return in_array($value_1, $value_2)
      ? !$this->_negated
      : $this->_negated;
  }


  public function getFormatDb(?bool $with_quote = true) : string {
    return $with_quote
      ? Zend_Db_Table_Abstract::getDefaultAdapter()
      ->quoteInto($this->_clauseFormatDb(),
                  $this->getValue(), null, null)
      : str_replace('?', implode(', ', $this->getValue()), $this->_clauseFormatDb());
  }


  protected function _clauseFormatDb() : string {
    return $this->_key->key($this->_identifier) . $this->_getOperator() . '(?)';
  }
}




class Storm_Query_ClauseIsNull extends Storm_Query_Clause {

  protected function _prepareValue() : self {
    $this->_value = null;

    return $this;
  }


  public function getFormatDb(?bool $with_quote = true) : string {
    return $this->_key->key($this->_identifier) . $this->_getOperator() . 'null';
  }


  protected function _getOperator() : string {
    return ' ' . $this->_operator . ($this->_negated ? static::CLAUSE_NOT : ' ');
  }
}




class Storm_Query_ClauseGreater extends Storm_Query_Clause {

  public function contains($value_1, $value_2) : bool {
    return $value_1 > $value_2;
  }
}




class Storm_Query_ClauseGreaterEqual extends Storm_Query_Clause {

  public function contains($value_1, $value_2) : bool {
    return $value_1 >= $value_2;
  }
}




class Storm_Query_ClauseLesser extends Storm_Query_Clause {

  public function contains($value_1, $value_2) : bool {
    return $value_1 < $value_2;
  }
}




class Storm_Query_ClauseLesserEqual extends Storm_Query_Clause {

  public function contains($value_1, $value_2) : bool {
    return $value_1 <= $value_2;
  }
}




class Storm_Query_ClauseLimit extends Storm_Query_Clause {

  protected function _prepareValue() : self {
    $this->_value = (string) $this->_value;

    return $this;
  }


  public function assemble($select) : self {
    if ($this->getValue())
      $select->limit($this->getValue(), 0);

    return $this;
  }


  public function getFormatDb(?bool $with_quote = true) : string {
    return '';
  }


  public function containAttibuteInVolatile(array $model) : bool {
    return true;
  }
}




class Storm_Query_ClauseLimitPage extends Storm_Query_Clause {

  protected function _prepareValue() : self {
    if ( ! $this->_value)
      return $this;

    if ( ! is_array($this->_value))
      $this->_value = [0, (int) $this->_value];

    $offset = (int) ($this->_value[0] ?? 1) - 1;
    if ($offset <= 0)
      $offset = 0;

    $limit = (int) ($this->_value[1] ?? 0);

    $this->_value = [$limit, $offset * $limit];

    return $this;
  }


  public function assemble($select) : self {
    if ($range = $this->getValue())
      $select->limit($range[0], $range[1]);

    return $this;
  }


  public function getFormatDb(?bool $with_quote = true) : string {
    return '';
  }


  public function containAttibuteInVolatile(array $model) : bool {
    return true;
  }
}




class Storm_Query_ClauseGroupBy extends Storm_Query_Clause {

  protected function _prepareValue() : self {
    $this->_value = (string)$this->_value;

    return $this;
  }


  public function assemble($select) : self {
    $identifier = '';
    if ($this->_identifier)
      $identifier = $this->_identifier . '.';

    $select->group(new Zend_Db_Expr($identifier
                                    . Zend_Db_Table_Abstract::getDefaultAdapter()
                                    ->quoteIdentifier($this->getValue(), true)));

    return $this;
  }


  public function getFormatDb(?bool $with_quote = true) : string {
    return '';
  }


  public function containAttibuteInVolatile(array $model) : bool {
    return true;
  }
}




class Storm_Query_ClauseWhere extends Storm_Query_Clause {

  protected function _prepareValue() : self {
    $this->_value = (string)$this->_value;

    return $this;
  }


  public function containAttibuteInVolatile(array $model) : bool {
    return true;
  }
}
