<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2022 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

*/


class Storm_Query_Condition {

  protected string $_parent_key;
  protected string $_key;
  protected string $_operator = Storm_Query_Clause::CLAUSE_EQUAL;
  protected bool $_negated = false;
  protected ?Storm_Query_Clause $_clause = null;
  protected Storm_Query_Abstract $_query;
  protected ?Storm_Query_Condition $_child_condition = null;

  protected function __construct(Storm_Query_Abstract $query, string $parent_key, string $key) {
    $this->_parent_key = $parent_key;
    $this->_key = $key;
    $this->_query = $query;
  }


  public static function newWith(Storm_Query_Abstract $query,
                                 string $parent_key,
                                 string $key) : self {
    if ($query instanceof Storm_Join_Inner)
      return new Storm_Query_ConditionInner($query, $parent_key, $key);
    if ($query instanceof Storm_Join_Left)
      return new Storm_Query_ConditionLeft($query, $parent_key, $key);
    if ($query instanceof Storm_Join_Right)
      return new Storm_Query_ConditionRight($query, $parent_key, $key);

    if ($query instanceof Storm_Query_In)
      return new Storm_Query_ConditionSubQueries($query, $parent_key, $key);

    if ($query instanceof Storm_Query_Exists)
      return new Storm_Query_ConditionSubQueries($query, $parent_key, $key);

    return new static($query, $parent_key, $key);
  }


  public function cloneCondition(Storm_Query_Abstract $query): self
  {
    $clone = static::newWith($query, $this->_parent_key, $this->_key);
    $clone->_operator = $this->_operator;
    $clone->_negated = $this->_negated;
    $clone->_clause = $this->_clause;

    if ($this->_child_condition)
      $clone->_child_condition = $this->_child_condition->cloneCondition($query);

    return $clone;
  }


  public function getParentKey() : string {
    return $this->_parent_key;
  }


  public function getQuery() : Storm_Query_Abstract {
    return $this->_query;
  }


  public function add(Storm_Query_Condition $child) : self {
    if ($this->_child_condition) {
      $this->_child_condition->add($child);

      return $this;
    }

    $this->_child_condition = $child;

    return $this;
  }


  public function beNotEqual() : self {
    $this->_negated = true;

    return $this;
  }


  public function beGreater() : self {
    $this->_operator = Storm_Query_Clause::CLAUSE_GREATER;

    return $this;
  }


  public function beGreaterEqual() : self {
    $this->_operator = Storm_Query_Clause::CLAUSE_GREATER_EQUAL;

    return $this;
  }


  public function beLesser() : self {
    $this->_operator = Storm_Query_Clause::CLAUSE_LESSER;

    return $this;
  }


  public function beLesserEqual() : self {
    $this->_operator = Storm_Query_Clause::CLAUSE_LESSER_EQUAL;

    return $this;
  }


  public function beIn() : self {
    $this->_operator = Storm_Query_Clause::CLAUSE_IN;

    return $this;
  }


  public function beNotIn() : self {
    $this->_operator = Storm_Query_Clause::CLAUSE_IN;
    $this->_negated = true;

    return $this;
  }


  public function isForSqlWhere() : bool {
    return Storm_Query_Clause::CLAUSE_IN !== $this->_operator;
  }


  public function sql() : array {
    if (!$this->_query->getParentHelper())
      return [];

    if ( ! $this->_child_condition)
      return [$this->_clause()->getFormatDb(false)];

    return [$this->_clause()->getFormatDb(false),
            ...$this->_child_condition->sql()];
  }


  protected function _clause() : Storm_Query_Clause {
    if ($this->_clause)
      return $this->_clause;

    $value = ($this->_query->getHelper()->getIdentifier()
              . '.' . Zend_Db_Table_Abstract::getDefaultAdapter()
              ->quoteIdentifier($this->_key, true));

    return Storm_Query_Clause::newFor(new Storm_Query_Key($this->_parent_key),
                                      $this->_operator,
                                      $value)
      ->setNegated($this->_negated)
      ->withHelperDo($this->_query->getParentHelper());
  }


  public function filterWith(Storm_Query_Instances $query_instances) : self {
    if (!$this->_query->getParentHelper())
      return $this;

    return $this->_filter($query_instances);
  }


  protected function _filter(Storm_Query_Instances $query_instances) : self {
    return $this;
  }


  public function conditionSelect(Storm_Query_Instance $parent_instance,
                                  Storm_Query_Instance $instance) : bool {
    $parent_value = $parent_instance->value($this->_query->getParentHelper()->getAlias(),
                                            $this->_parent_key);
    $value = $instance->value($this->_query->getHelper()->getAlias(), $this->_key);

    $contains = $this->_clause()->contains($parent_value, $value);

    if ( ! $contains || ! $this->_child_condition)
      return $contains;

    return $this->_child_condition->conditionSelect($parent_instance, $instance);
  }


  protected function _updateQueryInstances(Storm_Query_Instances $query_instances,
                                           Storm_Collection $filter) : self {
    $query_instances->addAll([$this->_query->getParentHelper()->getAlias(),
                              $this->_query->getHelper()->getAlias()],
                             $filter);

    return $this;
  }
}




class Storm_Query_ConditionInner extends Storm_Query_Condition {

  protected function _filter(Storm_Query_Instances $query_instances) : self {
    $alias_instances = $query_instances->instancesFor($this->_query->getHelper()->getAlias());
    $condition_filter = new Storm_Query_ConditionFilter($this, $alias_instances);

    $query_instances
      ->instancesFor($this->_query->getParentHelper()->getAlias())
      ->eachDo(fn($parent) => $condition_filter->updateCrossedWith($parent));

    return $this->_updateQueryInstances($query_instances, $condition_filter->getFilter());
  }
}




class Storm_Query_ConditionLeft extends Storm_Query_Condition {

  protected function _filter(Storm_Query_Instances $query_instances) : self {
    $alias_instances = $query_instances->instancesFor($this->_query->getHelper()->getAlias());
    $condition_filter = new Storm_Query_ConditionFilter($this, $alias_instances);

    $query_instances
      ->instancesFor($this->_query->getParentHelper()->getAlias())
      ->eachDo(fn($parent) => $condition_filter->updateOuterLeftWith($parent));

    return $this->_updateQueryInstances($query_instances, $condition_filter->getFilter());
  }
}




class Storm_Query_ConditionRight extends Storm_Query_Condition {

  protected function _filter(Storm_Query_Instances $query_instances) : self {
    $parent_instances = $query_instances
      ->instancesFor($this->_query->getParentHelper()->getAlias());
    $condition_filter = new Storm_Query_ConditionFilter($this, $parent_instances);

    $query_instances
      ->instancesFor($this->_query->getHelper()->getAlias())
      ->eachDo(fn($child) => $condition_filter->updateOuterRightWith($child));

    return $this->_updateQueryInstances($query_instances, $condition_filter->getFilter());
  }
}




class Storm_Query_ConditionSubQueries extends Storm_Query_ConditionInner {

  protected function _filter(Storm_Query_Instances $query_instances) : self {
    $alias_instances = $query_instances->instancesFor($this->_query->getHelper()->getAlias());
    $condition_filter = new Storm_Query_ConditionFilter($this, $alias_instances);

    $query_instances
      ->instancesFor($this->_query->getParentHelper()->getAlias())
      ->eachDo(fn($parent) => $condition_filter->updateAnyWith($parent));

    return $this->_updateQueryInstances($query_instances, $condition_filter->getFilter());
  }
}




class Storm_Query_ConditionFilter {

  protected Storm_Query_Condition $_condition;
  protected Storm_Collection $_child_instances;
  protected Storm_Collection $_filter;

  public function __construct(Storm_Query_Condition $condition,
                              Storm_Collection $child_instances) {
    $this->_condition = $condition;
    $this->_child_instances = $child_instances;
    $this->_filter = new Storm_Collection;
  }


  public function getFilter() : Storm_Collection {
    return $this->_filter;
  }


  public function updateCrossedWith(Storm_Query_Instance $parent) : self {
    $this->_child_instances
      ->select(fn($instance) => $this->_condition->conditionSelect($parent, $instance))
      ->eachDo(fn($instance) => $this->_filter->add((new Storm_Query_Instance)
                                                    ->addFrom($parent)
                                                    ->addFrom($instance)));

    return $this;
  }


  public function updateOuterLeftWith(Storm_Query_Instance $parent) : self {
    $selected = $this->_child_instances
      ->select(fn($instance) => $this->_condition->conditionSelect($parent, $instance));

    return $this->_withSelectedDo($selected, $parent);
  }


  public function updateOuterRightWith(Storm_Query_Instance $parent) : self {
    $selected = $this->_child_instances
      ->select(fn($instance) => $this->_condition->conditionSelect($instance, $parent));

    return $this->_withSelectedDo($selected, $parent);
  }


  protected function _withSelectedDo(Storm_Collection $selected,
                                     Storm_Query_Instance $parent) : self {
    if ($selected->isEmpty()) {
      $this->_filter->add((new Storm_Query_Instance)->addFrom($parent));

      return $this;
    }

    $selected
      ->eachDo(fn($instance) => $this->_filter->add((new Storm_Query_Instance)
                                                    ->addFrom($parent)
                                                    ->addFrom($instance)));

    return $this;
  }


  public function updateAnyWith(Storm_Query_Instance $parent): self
  {
    $exists = $this->_child_instances
      ->select(fn($instance) => $this->_condition->conditionSelect($parent, $instance));
    $found = ! $exists->isEmpty();

    if ($found && $this->_condition->getQuery()->isNegated())
      return $this;

    if ( ! $found && ! $this->_condition->getQuery()->isNegated())
      return $this;

    if ( ! $found) {
      $this->_filter->add((new Storm_Query_Instance)->addFrom($parent));

      return $this;
    }

    $exists->eachDo(fn($instance) => $this->_filter->add((new Storm_Query_Instance)
                                                         ->addFrom($parent)
                                                         ->addFrom($instance)));

    return $this;
  }
}
