<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2022 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

*/


class Storm_Query_Helper {

  protected Storm_Model_Loader $_loader;
  protected string $_alias;
  protected string $_table_name;
  protected string $_identifier;
  protected array $_select = [];
  protected array $_clauses_group_by = [];
  protected bool $_select_with_alias = false;
  protected bool $_is_distinct = false;
  protected bool $_all_select = true;
  protected bool $_is_join = false;
  protected bool $_is_sub = false;

  protected function __construct(Storm_Model_Loader $loader, ?string $alias = null) {
    $this->_loader = $loader;
    $this->_table_name = $loader->getTable()->getName();
    $this->_alias = $alias ?? $this->_table_name;
    $this->_identifier = Zend_Db_Table_Abstract::getDefaultAdapter()
      ->quoteIdentifier($this->_alias, true);
  }


  public static function newHelper(Storm_Model_Loader $loader, ?string $alias = null) : self {
    return new static($loader, $alias);
  }


  public static function newJoinHelper(Storm_Model_Loader $loader, ?string $alias = null) : self {
    $helper = new static($loader, $alias);
    $helper->_all_select = false;
    $helper->_is_join = true;

    return $helper;
  }


  public static function newSubHelper(Storm_Model_Loader $loader, ?string $alias = null) : self {
    $helper = new static($loader, $alias);
    $helper->_all_select = false;
    $helper->_is_sub = true;

    return $helper;
  }


  public function copyHelper(Storm_Query_Helper $helper, bool $copy_select): self
  {
    if ($helper->_is_join)
      $this->_all_select = $helper->_all_select;

    if ( ! $copy_select)
      return $this;

    $this->_select = [...$helper->_select];
    $this->_clauses_group_by = [...$helper->_clauses_group_by];

    $this->_is_distinct = $helper->_is_distinct;

    return $this;
  }


  public function getSelect() : array {
    return $this->_select;
  }


  public function computeErrors() : array {
    return [];
  }


  /**
   * @param $select : array<string|Storm_Query_Select> | string | Storm_Query_Select
   */
  public function addSelect($select) : self {
    if ( ! is_array($select))
      $select = [$select];

    foreach ($select as $key => $column)
      $this->_select [] = $this->_selectKey($column, $key);

    return $this;
  }


  public function addCountAll() : self {
    $this->_select = [];

    return $this->addSelect(['numberof' => Storm_Query_Key::count()]);
  }


  protected function _selectKey($column, $key) : Storm_Query_Key {
    if ( ! ($column instanceof Storm_Query_Key))
      $column = new Storm_Query_Key($column);

    if (is_string($key))
      $column->setAlias($key);

    return $column;
  }


  public function having(string $column, Storm_Query_Clause $clause) : self {
    $select = (new Storm_Collection($this->_select))
      ->detect(fn($curr_select) => $column === $curr_select->alias());

    if ($select)
      $select->addClause($clause);

    return $this;
  }


  public function hasSelect() : bool {
    return ($this->_all_select || $this->_select) && ! $this->_is_sub;
  }


  public function countSelect() : int {
    return count($this->_select);
  }


  public function beDistinct() : self {
    $this->_is_distinct = true;

    return $this;
  }


  public function beWithAlias() : self {
    if ($this->_select && ! $this->_all_select)
      $this->_select_with_alias = true;

    return $this;
  }


  public function setAllSelect(bool $all_select) : self {
    $this->_all_select = $all_select;

    return $this;
  }


  public function getAlias() : string {
    return $this->_alias;
  }


  public function getIdentifier() : string {
    return $this->_identifier;
  }


  public function getLoader() : Storm_Model_Loader {
    return $this->_loader;
  }


  public function getTableName() : string {
    return $this->_table_name;
  }


  public function setGroupBy(string $group_by) : self {
    $this->_clauses_group_by [] = Storm_Query_Clause::group($group_by)->withHelperDo($this);

    return $this;
  }


  public function isDistinct() : bool {
    return $this->_is_distinct;
  }


  public function isGroupBy() : bool {
    return (bool) count($this->_clauses_group_by);
  }


  public function groupBys() : array {
    return $this->_clauses_group_by;
  }


  public function assembleGroupBy(Zend_Db_Table_Select $select) : self {
    foreach ($this->_clauses_group_by as $group_by)
      $group_by->assemble($select);

    return $this;
  }


  public function loaderForFetch(): Storm_Model_BaseLoader
  {
    return $this->loaderIsRow()
      ? Storm_Model_Row::getLoader()
      : $this->getLoader();
  }


  public function loaderIsRow(): bool
  {
    return $this->_select || $this->_is_join;
  }


  /**
   * @return = null | string | array
   */
  public function select() {
    if ( ! $this->_select && ! $this->_all_select)
      return null;

    if ( ! $this->_select && $this->_all_select)
      return Zend_Db_Select::SQL_WILDCARD;

    $selects = [];
    foreach ($this->_select as $select)
      $selects = $this->_aliasFor($select, $selects, $select->key($this->_identifier));

    return $selects;
  }


  public function selectDefault() : array {
    if ( ! $this->_select || $this->_all_select)
      return [];

    $selects = [];
    foreach ($this->_select as $select)
      $selects = $this->_aliasFor($select, $selects, null);

    return $selects;
  }


  protected function _aliasFor(Storm_Query_Key $select, array $selects, $value) : array {
    $alias = $select->alias($this->aliasForSelect());

    if ($alias) {
      $selects [$alias] = $value;
      return $selects;
    }

    $selects [] = $value;
    return $selects;
  }


  public function withAttribGetSelect(string $attrib) : ?Storm_Query_Key {
    if ( ! $this->_select && ! $this->_all_select)
      return null;

    if ( ! $this->_select && $this->_all_select)
      return new Storm_Query_Key($attrib);

    foreach ($this->_select as $select)
      if ($attrib === $select->getKey())
        return $select;

    return null;
  }


  public function aliasForSelect() : ?string {
    return $this->_select_with_alias
      ? ($this->_alias . '_')
      : null;
  }
}
