<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2022 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

*/


class Storm_Query_Instances {

  const INSTANCE_SEPARATOR = '@@';
  protected array $_instances = [];

  public function __construct(Storm_Query_Abstract $query) {
    $this->_init($query);
  }


  protected function _init(Storm_Query_Abstract $query) : self {
    foreach ($query->instances() as $alias => $instances)
      $this->_instances [$alias] = $this->_addInstance($alias, $instances, $query);

    return $this;
  }


  public function addAll(array $searchs_alias, Storm_Collection $collection) : self {
    $all_alias = [];
    foreach ($searchs_alias as $search_alias)
      $all_alias [] = $this->_aliasKey($search_alias);

    $final_alias = [];
    foreach ($all_alias as $alias) {
      unset($this->_instances[$alias]);
      $final_alias = [...$final_alias, ...explode(static::INSTANCE_SEPARATOR, $alias)];
    }

    $this->_instances [implode(static::INSTANCE_SEPARATOR,
                               array_unique(array_filter($final_alias)))] = $collection;
    return $this;
  }


  public function instancesFor(string $search_alias) : Storm_Collection {
    return ($alias = $this->_aliasKey($search_alias))
      ? $this->_instances[$alias]
      : new Storm_Collection;
  }


  public function resultFor(Storm_Query_Abstract $query): array
  {
    $default_instance = $query->getJoins() ? $this->_defaultInstance($query) : [];

    $instances = $this->instancesFor($query->getHelper()->getAlias())
                      ->injectInto(new Storm_Collection,
                                   fn($instances, $instance) => $this->_linkedInstance($instances,
                                                                                       $instance,
                                                                                       $query));
    if ($query->hasOrders())
      $instances->uasort(fn($a, $b) => $query->orderVolatile($a, $b));

    $results = [];
    foreach ($instances->getArrayCopy() as $instance)
      $results[] = $instance->result($query, $default_instance);

    return $results;
  }


  protected function _linkedInstance(Storm_Collection $instances,
                                     Storm_Query_Instance $instance,
                                     Storm_Query_Abstract $query) : Storm_Collection {
    if ($found = $instances
        ->detect(fn($existing_instance) => $existing_instance->shouldLinkWith($instance,
                                                                              $query))) {
      $found->addLinkedAndSortBy($instance, $query);

      return $instances;
    }

    $instances->add($instance);

    return $instances;
  }


  protected function _defaultInstance(Storm_Query_Abstract $query) : array {
    $default = $query->getHelper()->selectDefault();

    foreach ($query->getJoins() as $join)
      $default = array_merge($default, $this->_defaultInstance($join));

    return $default;
  }


  protected function _addInstance(string $alias,
                                  array $instances,
                                  Storm_Query_Abstract $query) : Storm_Collection {
    $collection = new Storm_Collection;

    foreach ($instances as $instance)
      $collection->add((new Storm_Query_Instance)->addWith($alias, $instance));

    return $collection;
  }


  protected function _aliasKey(string $search_alias) : string {
    foreach (array_keys($this->_instances) as $alias)
      if (in_array($search_alias, explode(static::INSTANCE_SEPARATOR, $alias)))
        return $alias;

    return '';
  }
}
