<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2024 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  gof this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

*/


class Storm_Query_Terms extends Storm_Query_Terms_Abstract
{

  const OPERATOR_STRICT = '+';
  const OPERATOR_NOT = '-';
  const SEPARATOR = ' ';

  protected ?Storm_Query_Terms $_next_terms = null;
  protected bool $_exact = false;
  protected bool $_strict = false;
  protected bool $_negated = false;
  protected array $_match_terms = [];

  protected function __construct(array $match_terms)
  {
    foreach ($match_terms as $value)
      $this->_addTerm($value);
  }


  public static function and(array $match_terms): self
  {
    return (new static($match_terms))->_beStrict();
  }


  public static function or(array $match_terms): self
  {
    return new static($match_terms);
  }


  public static function not(array $match_terms): self
  {
    return (new static($match_terms))->_beNegated();
  }


  public static function andExact(string $value): self
  {
    return static::and([$value])->_beExact();
  }


  public static function orExact(string $value): self
  {
    return static::or([$value])->_beExact();
  }


  public static function notExact(string $value): self
  {
    return static::not([$value])->_beExact();
  }


  public function clone(): self
  {
    $clone = new static($this->_match_terms);

    $clone->_exact = $this->_exact;
    $clone->_strict = $this->_strict;
    $clone->_negated = $this->_negated;

    if ($this->_next_terms)
      $clone->_next_terms = $this->_next_terms->clone();

    return $clone;
  }


  protected function _addTerm($value): self
  {
    if ($value instanceof Storm_Query_Terms)
      return $this->_addTerms($value);

    if ($value instanceof Storm_Query_Terms_Term)
    {
      $this->_match_terms [] = $value;

      return $this;
    }

    if (is_int($value))
      $value = (string) $value;

    $this->_match_terms [] = new Storm_Query_Terms_Term($value);

    return $this;
  }


  protected function _addTerms(Storm_Query_Terms $terms): self
  {
    if ($terms->isOr() && ! $terms->_exact && $this->_match_terms !== [])
    {
      $this->_match_terms = [$this->clone(), $terms];

      return $this->_becomeOr();
    }

    $this->_match_terms [] = $terms;

    return $this;
  }


  protected function _becomeOr(): self
  {
    $this->_negated = false;
    $this->_strict = false;
    $this->_exact = false;
    $this->_next_terms = null;

    return $this;
  }


  public function addTerms(Storm_Query_Terms $terms): self
  {
    if ($terms->isOr())
      return $this->_addTerm($terms);

    if ( ! $this->_next_terms)
    {
      $this->_next_terms = $terms;

      return $this;
    }

    $this->_next_terms->addTerms($terms);

    return $this;
  }


  public function isOr(): bool
  {
    return ! $this->_strict && ! $this->_negated;
  }


  public function isNegated(): bool
  {
    return $this->_negated;
  }


  public function operator(): string
  {
    if ($this->_strict)
      return static::OPERATOR_STRICT;

    return $this->_negated
      ? static::OPERATOR_NOT
      : '';
  }


  protected function _beExact(): self
  {
    $this->_exact = true;

    return $this;
  }


  protected function _beStrict(): self
  {
    $this->_strict = true;
    $this->_negated = false;

    return $this;
  }


  protected function _beNegated(): self
  {
    $this->_negated = true;
    $this->_strict = false;

    return $this;
  }


  public function getFormatDb(): string
  {
    $surround = $this->isOr() && ! $this->_exact && count($this->_match_terms) > 1;

    $formats = ((new Storm_Collection($this->_match_terms))
                ->collect(fn($term) => $this->_surround($term->getFormatDb(),
                                                        $surround && ! $term->isOr()))
                ->getArrayCopy());

    $format = $this->_formatWithOperator(implode(static::SEPARATOR,
                                                 array_filter($formats,
                                                              fn($val) => '' !== $val)));

    return $this->_next_terms
      ? $this->_concat($format, $this->_next_terms->getFormatDb())
      : $format;
  }


  protected function _formatWithOperator(string $format): string
  {
    if ('' === $format)
      return '';

    $format = $this->_surround($format, $this->_mustSurround());

    $operator = $this->operator();
    if ($operator === ($format[0] ?? ''))
      $operator = '';

    return $operator . $format;
  }


  protected function _mustSurround(): bool
  {
    if ($this->_exact)
      return true;

    if ($this->isOr())
      return false;

    if (1 === count($this->_match_terms))
    {
      $terms = $this->_match_terms[0];

      return ($terms instanceof Storm_Query_Terms) && $this->operator() !== $terms->operator();
    }

    foreach ($this->_match_terms as $terms)
      if ($this->operator() !== $terms->operator())
        return true;

    return false;
  }


  protected function _surround(string $format, bool $surround): string
  {
    $prefix = '(';
    $suffix = ')';

    if ($this->_exact)
      $prefix = $suffix = '"';

    return ('' === $format) || ! $surround
      ? $format
      : $prefix . $format . $suffix;
  }


  public function containVolatile(string $contents, ?bool $contains = null): bool
  {
    if ($this->_isTermsHaveOnlyNegatedOperator())
      return false;

    $contain = false;
    $for_exact = $this->_exact && $this->_match_terms !== [];
    $match_terms = $this->_exact ? [] : $this->_match_terms;

    if ($for_exact)
      $contain = $this->_termForExact()->containVolatile($contents);

    foreach ($match_terms as $term)
      $contain |= $term->containVolatile($contents);

    if ($this->_strict && ! $contain)
      return false;

    if ($this->_negated && $contain)
      return false;

    if (null === $contains)
      $contains = $this->_strict || $this->_negated;

    $contains = $this->_containWithOperator($contains, $contain);

    return $this->_next_terms
      ? $this->_next_terms->containVolatile($contents, $contains)
      : $contains;
  }


  protected function _containWithOperator(bool $contains, bool $contain): bool
  {
    if ($this->_strict)
      return $contains && $contain;

    return $this->_negated
      ? $contains && ! $contain
      : $contains || $contain;
  }


  public function getCompareValue(string $contents, int $compares = 0): int
  {
    if ($this->_isTermsHaveOnlyNegatedOperator())
      return 0;

    $compare = 0;
    $for_exact = $this->_exact && $this->_match_terms !== [];
    $match_terms = $this->_exact ? [] : $this->_match_terms;

    if ($for_exact)
      $compare = $this->_termForExact()->getCompareValue($contents);

    foreach ($match_terms as $term)
      $compare += $term->getCompareValue($contents);

    if ($this->_strict && $compare <= 0)
      return 0;

    if ($this->_negated && $compare > 0)
      return 0;

    $compares += $this->_compareWithOperator($compare);

    return $this->_next_terms
      ? $this->_next_terms->getCompareValue($contents, $compares)
      : $compares;
  }


  protected function _compareWithOperator(int $compare): int
  {
    if ($this->_strict)
      return max($compare, 0);

    return $this->_negated
      ? ($compare < 0 ? -$compare : 0)
      : (max($compare, 0));
  }


  protected function _termForExact(): Storm_Query_Terms_Term
  {
    $value = '';

    foreach ($this->_match_terms as $term)
      $value = $this->_concat($value, $term->getFormatDb());

    return new Storm_Query_Terms_Term($value);
  }


  protected function _concat(string $prefix, string $suffix): string
  {
    if ('' === $prefix)
      return $suffix;

    return '' === $suffix
      ? $prefix
      : $prefix . static::SEPARATOR . $suffix;
  }


  // MariaDb 10 : only operator negated => 0 result
  protected function _isTermsHaveOnlyNegatedOperator(): bool
  {
    foreach ($this->_match_terms as $terms)
      if ( ! $terms->isNegated())
        return false;

    return true;
  }
}
