<?php
/*
STORM is under the MIT License (MIT)

Copyright (c) 2010-2022 Agence Française Informatique http://www.afi-sa.fr

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/


class Storm_Test_QueriesObserver {

  protected Storm_Collection $_queries;
  protected Storm_Set $_scopes;

  public function __construct() {
    $this->_queries = new Storm_Collection;
    $this->_scopes = new Storm_Set;
  }


  public function addScope(string $scope) : self {
    $this->_scopes->add($scope);

    return $this;
  }


  public function __invoke(Storm_Event_Abstract $event) : self {
    if ($this->_scopes->isEmpty() || ! $event->isQueryEvent())
      return $this;

    ob_start();
    debug_print_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS);
    $stack = ob_get_contents();
    ob_end_clean();

    if ($this->_scopes->detect(fn($scope) => false !== strpos($stack, (string) $scope)))
      $this->_addQuery($event->getQuery());

    return $this;
  }


  public function detectWith(string $sql, ?int $count = null) : ?Storm_Test_QueriesObserverSql {
    return $this->_queries
      ->detect(fn($observer) => $sql === $observer->getSql()
                                && (null === $count || $count === $observer->getCount()));
  }


  public function compareSize(array $sqls) : bool {
    $queries = new Storm_Collection;

    foreach ($sqls as $sql)
      if ($query = $this->detectWith($sql))
        $queries->add($query);

    $observers_count = $this->_queries->count();
    $queries_count = $queries->count();

    return ($queries_count === $observers_count) && ($queries_count === count($sqls));
  }


  public function isEmpty() : bool {
    return $this->_queries->isEmpty();
  }


  public function queriesMessages() : array {
    return $this->_queries
      ->collect(fn($observer) => $observer->getSql()
                . sprintf(' / Called %d time(s)', $observer->getCount()))
      ->getArrayCopy();
  }


  protected function _addQuery(Storm_Query_Abstract $query) : self {
    $query_observer = new Storm_Test_QueriesObserverSql($query);

    $old_observer = $this->_queries
      ->detect(fn($observer) => $observer->getKey() === $query_observer->getKey());

    if ( ! $old_observer) {
      $this->_queries->add($query_observer);
      $old_observer = $query_observer;
    }

    $old_observer->increment();
    return $this;
  }
}




class Storm_Test_QueriesObserverSql {

  protected string $_sql = '';
  protected string $_key = '';
  protected int $_count = 0;

  public function __construct(Storm_Query_Abstract $query) {
    $this->_sql = ($errors = $query->computeErrors())
      ? implode(', ', $errors)
      : $query->assembleDb();

    $this->_key = $query->uniqKey($this->_sql);
  }


  public function getKey() : string {
    return $this->_key;
  }


  public function getSql() : string {
    return $this->_sql;
  }


  public function getCount() : int {
    return $this->_count;
  }


  public function increment() : self {
    $this->_count++;

    return $this;
  }
}
