<?php
/*
STORM is under the MIT License (MIT)

Copyright (c) 2010-2024 Agence Française Informatique http://www.afi-sa.fr

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/


class Storm_Test_ClearCacheTest extends Storm_Test_ModelTestCase
{

  protected ClearCache_User $_user_1;
  protected ClearCache_User $_user_2;
  protected ClearCache_User $_user_3;

  public function setUp(): void
  {
    parent::setUp();

    Storm_Model_Abstract::unsetLoaders();
    Storm_Model_Loader::resetCache();
    Storm_Model_Loader::defaultToVolatile();

    $this->_user_1 = $this->fixture(ClearCache_User::class,
                                    ['id' => 1,
                                     'name' => 'user 1']);

    $this->_user_2 = ClearCache_User::newInstance(['name' => 'user 2']);
    $this->_user_2->assertSave();

    $this->_user_3 = ClearCache_User::newInstanceWithId(3, ['name' => 'user 3']);
  }


  public function tearDown(): void
  {
    Storm_Model_Abstract::unsetLoaders();
    Storm_Model_Loader::resetCache();
    Storm_Model_Loader::defaultToDb();
    parent::tearDown();
  }


  /** @test */
  public function originalStateShouldBeHasExpected()
  {
    $this->assertTrue($this->_user_1->hasBeenSaved());
    $this->assertFalse($this->_user_1->hasBeenUpdated());
    $this->assertFalse($this->_user_1->hasBeenDeleted());
    $this->assertFalse($this->_user_1->hasBeenLoaded());
    $this->assertTrue($this->_user_1->isFixture());

    $this->assertTrue($this->_user_2->hasBeenSaved());
    $this->assertFalse($this->_user_2->hasBeenUpdated());
    $this->assertFalse($this->_user_2->hasBeenDeleted());
    $this->assertFalse($this->_user_2->hasBeenLoaded());
    $this->assertFalse($this->_user_2->isFixture());

    $this->assertFalse($this->_user_3->hasBeenSaved());
    $this->assertFalse($this->_user_3->hasBeenUpdated());
    $this->assertFalse($this->_user_3->hasBeenDeleted());
    $this->assertFalse($this->_user_3->hasBeenLoaded());
    $this->assertFalse($this->_user_3->isFixture());

    $this->assertCount(2, ClearCache_User::getLoadedInstances());
  }


  /** @test */
  public function withDeleteUser1AndClearCacheShouldExistInLoaderInstanceWithStateDeleted()
  {
    $user_1 = ClearCache_User::find(1);
    $user_1->delete();
    ClearCache_User::clearCache();

    $this->assertNull(ClearCache_User::find(1));
    $this->assertNotNull(ClearCache_User::find(2));
    $this->assertNull(ClearCache_User::find(3));

    $this->assertTrue($user_1->hasBeenDeleted());
    $this->assertTrue($user_1->hasBeenLoaded());
    $this->assertFalse($user_1->hasBeenUpdated());
    $this->assertTrue($user_1->isFixture());

    $this->assertFalse($this->_user_2->hasBeenDeleted());
    $this->assertFalse($this->_user_2->hasBeenUpdated());

    $this->assertFalse($this->_user_3->hasBeenDeleted());
    $this->assertFalse($this->_user_3->hasBeenUpdated());

    $this->assertCount(1, ClearCache_User::getLoadedInstances());
  }


  /** @test */
  public function withDeleteUser2ShouldNotExistInLoaderInstanceWithStateDeleted()
  {
    $user_2 = ClearCache_User::find(2);
    $user_2->delete();

    $this->assertNotNull(ClearCache_User::find(1));
    $this->assertNull(ClearCache_User::find(2));
    $this->assertNull(ClearCache_User::find(3));

    $this->assertFalse($this->_user_1->hasBeenDeleted());
    $this->assertFalse($this->_user_1->hasBeenUpdated());

    $this->assertTrue($user_2->hasBeenDeleted());
    $this->assertTrue($user_2->hasBeenLoaded());
    $this->assertFalse($user_2->hasBeenUpdated());
    $this->assertFalse($user_2->isFixture());

    $this->assertFalse($this->_user_3->hasBeenDeleted());
    $this->assertFalse($this->_user_3->hasBeenUpdated());

    $this->assertCount(1, ClearCache_User::getLoadedInstances());
  }


  /** @test */
  public function withClearCacheUser2ShouldNotExistInLoaderInstance()
  {
    $user_1 = ClearCache_User::find(1);
    $user_2 = ClearCache_User::find(2);
    ClearCache_User::clearCache();

    $this->assertNotNull(ClearCache_User::find(1));
    $this->assertNotNull(ClearCache_User::find(2));

    $this->assertTrue($user_1->hasBeenSaved());
    $this->assertFalse($user_1->hasBeenUpdated());
    $this->assertFalse($user_1->hasBeenDeleted());
    $this->assertTrue($user_1->hasBeenLoaded());
    $this->assertTrue($user_1->isFixture());

    $this->assertTrue($user_2->hasBeenSaved());
    $this->assertFalse($user_2->hasBeenUpdated());
    $this->assertFalse($user_2->hasBeenDeleted());
    $this->assertTrue($user_2->hasBeenLoaded());
    $this->assertFalse($user_2->isFixture());

    $this->assertCount(2, ClearCache_User::getLoadedInstances());
  }


  /** @test */
  public function updateUser1WithSaveShouldBeHasExpected()
  {
    $this->_user_1->setName('user 1 updated')->assertSave();

    $this->assertTrue($this->_user_1->hasBeenUpdated());

    $this->assertEquals('user 1 updated', ClearCache_User::find(1)->getName());
  }


  /** @test */
  public function updateUser1WithQueryShouldBeHasExpected()
  {
    ClearCache_User::updateQuery()
      ->addSet('name', 'user 1 updated')
      ->eq('id', 1)
      ->updateAll();

    $this->assertTrue($this->_user_1->hasBeenUpdated());

    $this->assertEquals('user 1 updated', ClearCache_User::find(1)->getName());
  }


  /** @test */
  public function onFindFirstByShouldBeWithStatusLoaded()
  {
    $user_1 = ClearCache_User::findFirstBy(['name' => 'user 1']);
    $this->assertTrue($user_1->hasBeenLoaded());
    $this->assertFalse($this->_user_2->hasBeenLoaded());
    $this->assertFalse($this->_user_3->hasBeenLoaded());
  }


  /** @test */
  public function onQueryFetchFirstShouldBeWithStatusLoaded()
  {
    $user_1 = ClearCache_User::query()
      ->eq('name', 'user 1')
      ->fetchFirst();
    $this->assertTrue($user_1->hasBeenLoaded());
    $this->assertFalse($this->_user_2->hasBeenLoaded());
    $this->assertFalse($this->_user_3->hasBeenLoaded());
  }


  /** @test */
  public function onFindAllShouldAllUsersStatusLoaded()
  {
    $users = ClearCache_User::findAll();
    $this->assertEquals([true, true],
                        array_map(fn($model) => $model->hasBeenLoaded(),
                                  $users));
  }


  /** @test */
  public function onFindAllByShouldAllUsersStatusLoaded()
  {
    $users = ClearCache_User::findAllBy(['name' => ['user 1', 'user 2']]);
    $this->assertEquals([true, true],
                        array_map(fn($model) => $model->hasBeenLoaded(),
                                  $users));
  }


  /** @test */
  public function onQueryFetchAllShouldAllUsersStatusLoaded()
  {
    $users = ClearCache_User::query()
      ->in('name', ['user 1', 'user 2'])
      ->fetchAll();
    $this->assertEquals([true, true],
                        array_map(fn($model) => $model->hasBeenLoaded(),
                                  $users));
  }
}




class ClearCache_User extends Storm_Model_Abstract
{

  protected $_table_name = 'users';
}
