<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2024 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
*/

require_once __DIR__ . '/Fixtures.php';

class Storm_Test_Query_LoaderVolatileClauseOrderTest extends Storm_Test_ModelTestCase
{

  protected array $_storm_scopes = ['LoaderVolatileClauseOrderTest'];

  public function setUp(): void
  {
    parent::setUp();

    Storm_Model_Loader::defaultToVolatile();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 100,
                    'login' => 'user_admin',
                    'level' => 'invite admin redacteur',
                    'foo' => 'first second third',
                    'order1' => 'abcd',
                    'order2' => 2,
                    'order3' => 'defg',
                    'order4' => 'weburl',
                    'order5' => '234']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 101,
                    'login' => 'user_administrateur',
                    'level' => 'administrateur',
                    'foo' => 'third second forth',
                    'order1' => 'abcdef',
                    'order2' => 3,
                    'order3' => 'defg',
                    'order4' => 'no',
                    'order5' => '12']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 102,
                    'login' => 'user_invite',
                    'level' => 'invite',
                    'foo' => 'forth fifth seven',
                    'order1' => 'cdef',
                    'order2' => 1,
                    'order3' => 'klmn',
                    'order4' => '',
                    'order5' => '111']);
  }


  public function tearDown(): void
  {
    Storm_Model_Loader::defaultToDb();
    Storm_Model_Loader::resetCache();
    parent::tearDown();
  }


  /** @test */
  public function with_Order1Desc_ShouldAnswersInOrdered()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc('order1');

    $this->assertEquals(['user_invite', 'user_administrateur', 'user_admin'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY `users`.`order1` DESC');
  }


  /** @test */
  public function with_Order3Asc_Order2Desc_ShouldAnswersOrderedUser()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order('order3')
      ->order_desc('order2');

    $this->assertEquals(['user_administrateur', 'user_admin', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY `users`.`order3` ASC, `users`.`order2` DESC');
  }


  /** @test */
  public function with_Order3Desc_Order2Desc_ShouldAnswersOrderedUser()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc('order3')
      ->order_desc('order2');

    $this->assertEquals(['user_invite', 'user_administrateur', 'user_admin'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY `users`.`order3` DESC, `users`.`order2` DESC');
  }


  /** @test */
  public function withMatch_Level_Administrateur_ShouldAnswersOrderedUser()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order((new Storm_Query_MatchRating('level'))
              ->against_or('administrateur'));

    $this->assertEquals(['user_admin', 'user_invite', 'user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY MATCH(`users`.`level`) AGAINST(\'administrateur\') ASC');
  }


  /** @test */
  public function withMatch_Level_Administrateur_Redacteur_Desc_ShouldAnswersOrderedUser()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc((new Storm_Query_MatchRating('level'))
                   ->against_or('administrateur redacteur'));

    $this->assertEquals(['user_administrateur', 'user_admin', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY MATCH(`users`.`level`) AGAINST(\'administrateur redacteur\') DESC');
  }


  /** @test */
  public function withMatchInBooleanMode_Level_Administrateur_Redacteur_Desc_ShouldAnswersOrderedUser()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc((new Storm_Query_MatchBoolean('level'))
                   ->against_or('administrateur redacteur'));

    $this->assertEquals(['user_admin', 'user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY MATCH(`users`.`level`) AGAINST(\'administrateur redacteur\' IN BOOLEAN MODE) DESC');
  }


  /** @test */
  public function withMatch_FooDescExacteInBooleanMode_Forth_Fifth_Level_Administrateur_Redacteur_ShouldAnswersOrderedUser()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc((new Storm_Query_MatchBoolean('foo'))
                   ->exact_or('forth fifth'))
      ->order((new Storm_Query_MatchRating('level'))
              ->against_or('administrateur redacteur'));

    $this->assertEquals(['user_invite', 'user_admin', 'user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY MATCH(`users`.`foo`) AGAINST(\'"forth fifth"\' IN BOOLEAN MODE) DESC, MATCH(`users`.`level`) AGAINST(\'administrateur redacteur\') ASC');
  }


  /** @test */
  public function withMatch_FooLevelStrict_InviteRedacteur_SecondThird_ShouldAnswersOrderedUser()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order((new Storm_Query_MatchBoolean('foo,level'))
              ->against_and('invite redacteur')
              ->against_and('second third'));

    $this->assertEquals([101, 102, 100],
                        array_map(fn($row) => $row->getId(), $query->fetchAll()));
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY MATCH(`users`.`foo`, `users`.`level`) AGAINST(\'+(invite redacteur) +(second third)\' IN BOOLEAN MODE) ASC');
  }


  /** @test */
  public function selectUsersGroupByOrder3ShouldAnswersUserAdminAndUserInvite()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->group('order3');

    $this->assertEquals(['user_admin', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` GROUP BY `users`.`order3`');
  }


  /** @test */
  public function withSelectShouldAnswersWithRowModel()
  {
    $results = Storm_Test_LoaderQueryUser::query()
      ->select(['login', 'level'])
      ->fetchAll();

    $user = Storm_Test_LoaderQueryUser::find(100);
    $this->assertEquals(['login' => $user->getLogin(),
                         'level' => $user->getLevel()],
                        $results[0]->toArray());
    $this->assertTrue($results[0] instanceof Storm_Model_Row);

    $user = Storm_Test_LoaderQueryUser::find(101);
    $this->assertEquals(['login' => $user->getLogin(),
                         'level' => $user->getLevel()],
                        $results[1]->toArray());
    $this->assertTrue($results[1] instanceof Storm_Model_Row);

    $user = Storm_Test_LoaderQueryUser::find(102);
    $this->assertEquals(['login' => $user->getLogin(),
                         'level' => $user->getLevel()],
                        $results[2]->toArray());
    $this->assertTrue($results[2] instanceof Storm_Model_Row);
    $this->assertSql('SELECT `users`.`login`, `users`.`level` FROM `users`');
  }


  /** @test */
  public function with_Order4WithNoValueDesc_ShouldAnswersOrderedUser_101_100_102()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc(Storm_Query_Clause::orderValue('order4', 'no'));

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSqlCount('SELECT `users`.* FROM `users` ORDER BY `users`.`order4` = \'no\' DESC', 1);
  }


  /** @test */
  public function with_Order1_ModeField_ShouldAnswersOrderedUser_101_102_100()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order(Storm_Query_Clause::orderField('order1', ['cdef', 'abcd']));

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(102),
                         Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY FIELD(`users`.`order1`, \'cdef\', \'abcd\') ASC');
  }


  /** @test */
  public function with_Order5_ModeCastInt_ShouldAnswersOrderedUser_101_102_100()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order(Storm_Query_Clause::orderCast('order5'));

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(102),
                         Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY CAST(`users`.`order5` AS SIGNED INTEGER) ASC');
  }


  /** @test */
  public function withOrderRandShouldReturnSqlWithRandCommandAndUseCallbackForVolatileOrder()
  {
    Storm_Test_LoaderQueryUser::randCallback(fn($a, $b) => $a->value('users', 'foo') <=> $b->value('users', 'foo'));
    $query = Storm_Test_LoaderQueryUser::query()
      ->rand();

    $this->assertEquals([100, 102, 101], array_map(fn($row) => $row->getId(),
                                                   $query->fetchAll()));
    $this->assertSql("SELECT `users`.* FROM `users` ORDER BY RAND()");
  }


  /** @test */
  public function withOrderRandWithoutCallbackShouldReturnSqlWithRandCommand()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->rand()
      ->fetchAll();

    $this->assertSql("SELECT `users`.* FROM `users` ORDER BY RAND()");
  }
}
