<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2024 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
*/

require_once __DIR__ . '/Fixtures.php';

class Storm_Test_Query_LoaderVolatileClauseWhereTest extends Storm_Test_ModelTestCase
{

  protected array $_storm_scopes = ['LoaderVolatileClauseWhereTest'];

  public function setUp(): void
  {
    parent::setUp();

    Storm_Model_Loader::defaultToVolatile();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 100,
                    'login' => 'user_admin',
                    'level' => 'invite admin redacteur',
                    'foo' => 'premier deuxieme troisieme',
                    'prefs' => null,
                    'sys' => '',
                    'date' => '2022-01-01']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 101,
                    'login' => 'user_administrateur',
                    'level' => 'administrateur',
                    'foo' => 'deuxieme redacteur',
                    'prefs' => '{"best":true}',
                    'sys' => null,
                    'date' => '2020-01-01']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 102,
                    'login' => 'user_invite',
                    'level' => 'invite',
                    'foo' => 'premier deuxieme',
                    'url' => 'https://my-server.org/page_1.html',
                    'website' => 'https://my-server.org/pomme+d\'api.php page2.php',
                    'sys' => 'content',
                    'date' => '2021-01-01']);
  }


  public function tearDown(): void
  {
    Storm_Model_Loader::defaultToDb();
    Storm_Model_Loader::resetCache();
    parent::tearDown();
  }


  /** @test */
  public function withLimit_0_ShouldReturnNothing()
  {
    $this->assertEmpty(Storm_Test_LoaderQueryUser::query()
                       ->limit(0)
                       ->fetchAll());
  }


  /** @test */
  public function withTwoFollowingFetchAllQueryShouldAnswersCorrectLoader()
  {
    $query = Storm_Test_LoaderQueryUser::query();

    $results = $query->fetchAll();
    $this->assertTrue($results[0] instanceof Storm_Test_LoaderQueryUser);

    $results = $query->fetchAll();
    $this->assertTrue($results[0] instanceof Storm_Test_LoaderQueryUser);
  }


  /** @test */
  public function withOnlyOneClauseInWrapCriteriaAndShouldNotStartWhereWithAnd()
  {
    $query = Storm_Test_LoaderQueryUser::query();

    $query->wrapWithAnd((new Storm_Query_Criteria)
                        ->eq('id', 100))
          ->fetchAll();

    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` = 100)');
  }


  /** @test */
  public function queryWrapShouldChangeOperator()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->eq('id', 100)
      ->gt('id', 101);

    $query->wrapWithOr((new Storm_Query_Criteria)
                       ->match((new Storm_Query_MatchRating('foo'))
                               ->against_or('premier')))
          ->fetchAll();

    $query->wrapWithAnd((new Storm_Query_Criteria)
                        ->start('sys', 'cont'))
          ->fetchAll();

    $query->append((new Storm_Query_Criteria)
                   ->like('foo', '%deux%')
                   ->not_eq('level', 'invite'))
          ->fetchAll();

    $this->assertSqlEquals(['SELECT `users`.* FROM `users` WHERE ((`users`.`id` = 100 AND `users`.`id` > 101) OR MATCH(`users`.`foo`) AGAINST(\'premier\'))',
                            'SELECT `users`.* FROM `users` WHERE (((`users`.`id` = 100 AND `users`.`id` > 101) OR MATCH(`users`.`foo`) AGAINST(\'premier\')) AND `users`.`sys` LIKE \'cont%\')',
                            'SELECT `users`.* FROM `users` WHERE (((`users`.`id` = 100 AND `users`.`id` > 101) OR MATCH(`users`.`foo`) AGAINST(\'premier\')) AND `users`.`sys` LIKE \'cont%\' AND (`users`.`foo` LIKE \'%deux%\' AND `users`.`level` != \'invite\'))']);
  }


  /** @test */
  public function queryCheckHasOrders()
  {
    $this->assertTrue(Storm_Test_LoaderQueryUser::query()
                      ->order('login')
                      ->hasOrders());

    $this->assertFalse(Storm_Test_LoaderQueryUser::query()
                       ->hasOrders());
  }


  /** @test */
  public function queryCheckHasClauses()
  {
    $this->assertTrue(Storm_Test_LoaderQueryUser::query()
                      ->eq('id', 100)
                      ->hasClauses());
    $this->assertTrue(Storm_Test_LoaderQueryUser::join('tab1')
                      ->inner(Storm_Test_LoaderQueryUser::inner('tab2')
                              ->on_eq('id', 'id')
                              ->gt('id', 100))
                      ->hasClauses());

    $this->assertFalse(Storm_Test_LoaderQueryUser::query()
                       ->hasClauses());
    $this->assertFalse(Storm_Test_LoaderQueryUser::query()
                       ->limit(5)
                       ->hasClauses());
    $this->assertFalse(Storm_Test_LoaderQueryUser::query()
                       ->limit_page(5, 50)
                       ->hasClauses());
    $this->assertFalse(Storm_Test_LoaderQueryUser::query()
                       ->group('id')
                       ->hasClauses());
    $this->assertFalse(Storm_Test_LoaderQueryUser::join('tab1')
                       ->inner(Storm_Test_LoaderQueryUser::inner('tab2')
                               ->on_eq('id', 'id'))
                       ->hasClauses());
  }


  /** @test */
  public function whereId_GT_100_ShouldAnswersUserAdministrateurAndUserInvite()
  {
    $this->assertEquals(['user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->gt('id', 100)
                                                    ->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` > 100)');
  }


  /** @test */
  public function whereId_GT_EQ_101_ShouldAnswersUserAdministrateurAndUserInvite()
  {
    $this->assertEquals(['user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->gt_eq('id', 101)
                                                    ->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` >= 101)');
  }


  /** @test */
  public function whereId_LT_102_ShouldAnswersUserAdminAndUserAdministrateur()
  {
    $this->assertEquals(['user_admin', 'user_administrateur'],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->lt('id', 102)
                                                    ->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSqlCount('SELECT `users`.* FROM `users` WHERE (`users`.`id` < 102)', 1);
  }


  /** @test */
  public function whereId_LT_EQ_101_ShouldAnswersUserAdminAndUserAdministrateur()
  {
    $this->assertEquals(['user_admin', 'user_administrateur'],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->lt_eq('id', 101)
                                                    ->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` <= 101)');
    $this->assertNotSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` > 101)');
  }


  /** @test */
  public function whereLogin_NOT_EQ_UserAdministrateurShouldReturnUserAdminAndUserInvite()
  {
    $this->assertEquals(['user_admin', 'user_invite'],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->not_eq('login', 'user_administrateur')
                                                    ->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`login` != \'user_administrateur\')');
  }


  /** @test */
  public function whereId_IN_100And102_AndFoo_NOT_LIKE_TroisieShouldReturnId_102()
  {
    $this->assertEquals([102],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->in('id', [100, 102])
                                                    ->not_like('foo', '%troisie%')
                                                    ->fetchAll()))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` IN (100, 102) AND `users`.`foo` NOT LIKE \'%troisie%\')');
  }


  /** @test */
  public function whereId_IN_100And101_AndFoo_START_DeuxShouldReturnId_101()
  {
    $this->assertEquals([101],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->in('id', [100, 101])
                                                    ->start('foo', 'deux')
                                                    ->fetchAll()))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` IN (100, 101) AND `users`.`foo` LIKE \'deux%\')');
  }


  /** @test */
  public function whereNotFoo_START_DeuxShouldReturnIds_100_102()
  {
    $this->assertEquals([100, 102],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->not_start('foo', 'deux')
                                                    ->fetchAll()))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`foo` NOT LIKE \'deux%\')');
  }


  /** @test */
  public function whereFoo_END_ActeurShouldReturnId_101()
  {
    $this->assertEquals([101],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->end('foo', 'acteur')
                                                    ->fetchAll()))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`foo` LIKE \'%acteur\')');
  }


  /** @test */
  public function whereNotFoo_END_ActeurShouldReturnId_100_102()
  {
    $this->assertEquals([100, 102],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->not_end('foo', 'acteur')
                                                    ->fetchAll()))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`foo` NOT LIKE \'%acteur\')');
  }


  /** @test */
  public function matchWithAgainstOn_0_ShouldReturnQueryWithAllAgainst()
  {
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 103,
                    'level' => '1 2 3']);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 104,
                    'level' => '9 0 8']);

    $result = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_or('1 0'))
      ->fetchAll();

    $this->assertEquals([103, 104],
                        (new Storm_Model_Collection($result))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'1 0\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function matchWithAndExactAndAgainstAndExactShouldReturnExpected()
  {
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 103,
                    'level' => '1 2 3']);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 104,
                    'level' => '1 3 2']);

    $result = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->exact_and('1 2')
              ->against_and('3')
              ->exact_and('2 3'))
      ->fetchAll();

    $this->assertEquals([103],
                        (new Storm_Model_Collection($result))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'+"1 2" +3 +"2 3"\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstAdminShouldAnswersOnlyUser_Admin()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_or('admin'));

    $this->assertEquals(['user_admin'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'admin\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withNotMatchLevelAgainstAdminShouldAnswersUsers_Administrateur_Invite()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->not_match((new Storm_Query_MatchBoolean('level'))
                  ->against_or(['admin']));

    $this->assertEquals(['user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (NOT MATCH(`users`.`level`) AGAINST(\'admin\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstAdminDeuxiemeShouldAnswersUser_AdminAndUser_AdministrateurAndUser_Invite()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_or(['admin', 'deuxieme']));

    $this->assertEquals(['user_admin', 'user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'admin deuxieme\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstNotAdminDeuxiemeShouldAnswersId_101()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_not(['admin', 'premier']));

    $this->assertEquals([101],
                        array_map(fn($row) => $row->getId(),
                                  $query->fetchAll()));
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'-(admin premier)\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstAdmin_StarShouldAnswersId_100_101()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_or('admin*'));

    $this->assertEquals([100, 101],
                        array_map(fn($row) => $row->getId(),
                                  $query->fetchAll()));
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'admin*\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstStar_ActeurShouldAnswersId_101()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('foo'))
              ->against_or('*acteur'));

    $this->assertEquals([101],
                        array_map(fn($row) => $row->getId(),
                                  $query->fetchAll()));
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`foo`) AGAINST(\'*acteur\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstStrictAdminRedacteur_DeuxiemeShouldAnswersUser_AdminAndUser_Administrateur()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_and(['admin', 'redacteur'])
              ->against_and(['deuxieme']));

    $this->assertEquals(['user_admin', 'user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'+(admin redacteur) +deuxieme\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelFooExactDeuxiemeRedacteurShouldAnswersOnlyUser_Administrateur()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->exact_or('deuxieme redacteur'));

    $this->assertEquals(['user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'"deuxieme redacteur"\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelFooDeuxiemeAndAgainstRedacteurAnswers_AllUsers()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_or('deuxieme')
              ->against_or('redacteur'));

    $this->assertEquals(['user_admin', 'user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'deuxieme redacteur\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelFooDeuxiemeAndAgainstStrictRedacteurAnswers_User_Admin_User_Administrateur()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_or('deuxieme')
              ->against_and('redacteur'));

    $this->assertEquals(['user_admin', 'user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'deuxieme +redacteur\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchWithMultipleTermsShouldReturn_Admin_Invite()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_and([Storm_Query_Terms::and(['admin']),
                             Storm_Query_Terms::and(['administrateurs', 'administration'])])
              ->against_or(['redacteurs', 'premier']));

    $logins = (new Storm_Model_Collection($query->fetchAll()))
      ->collect('login')
      ->getArrayCopy();
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'(+admin +(administrateurs administration)) redacteurs premier\' IN BOOLEAN MODE))');
    $this->assertEquals(['user_admin', 'user_invite'], $logins);
  }


  /** @test */
  public function withMatchWithMultipleTermsComplexShouldReturn_Admin_Invite()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_and([Storm_Query_Terms::or([Storm_Query_Terms::and(['admin', 'administration']),
                                                    'invite'])]));

    $logins = (new Storm_Model_Collection($query->fetchAll()))
      ->collect('login')
      ->getArrayCopy();
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'+((+(admin administration)) invite)\' IN BOOLEAN MODE))');
    $this->assertEquals(['user_admin', 'user_invite'], $logins);
  }


  /** @test */
  public function withOrMatchWithMultipleTermsAndComplexShouldReturnSql()
  {
    Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_or([Storm_Query_Terms::and([Storm_Query_Terms::and(['admin', 'administration']),
                                                    Storm_Query_Terms::and(['invite', 'premier'])]),
                            Storm_Query_Terms::and([Storm_Query_Terms::and(['admin2', 'administration2']),
                                                    Storm_Query_Terms::and(['invite2', 'premier2'])])]))
      ->fetchAll();

    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'(+(admin administration) +(invite premier)) (+(admin2 administration2) +(invite2 premier2))\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchWithAndTermsInsideOrShouldReturnSql()
  {
    Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_or([Storm_Query_Terms::and(['administrateurs', 'administration']),
                            Storm_Query_Terms::and(['invite', 'premier'])]))
      ->fetchAll();

    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'(+(administrateurs administration)) (+(invite premier))\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withSameDoubleOperatorShouldReturnSqlQueryWithOnlyOneOperator()
  {
    Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_and(Storm_Query_Terms::and(['admin'])))
      ->fetchAll();

    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'+admin\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withDifferentDoubleOperatorDistinctShouldReturnSqlQueryWithTwoOperator()
  {
    $result = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_or(Storm_Query_Terms::not(['admin'])))
      ->fetchAll();

    $this->assertCount(0, $result);
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'-admin\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withOrClause_LevelInviteOrFooLikePremierShouldAnswersUser_AdminAndUser_Invite()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->or((new Storm_Query_Criteria)
           ->eq('level', 'invite')
           ->like('foo', '%premier%'));

    $this->assertEquals(['user_admin', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE ((`users`.`level` = \'invite\' OR `users`.`foo` LIKE \'%premier%\'))');
  }


  /** @test */
  public function withAndClause_LevelInviteOrFooLikePremierShouldAnswersOnlyUser_Invite()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->and((new Storm_Query_Criteria)
            ->eq('level', 'invite')
            ->like('foo', '%premier%'));

    $this->assertEquals(['user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE ((`users`.`level` = \'invite\' AND `users`.`foo` LIKE \'%premier%\'))');
  }


  /** @test */
  public function urlLikeHttpsDoubleDotSlashSlashShouldAnswersOnlyUserWithSlashes()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->like('url', '%https://my-server.org%');

    $this->assertEquals(['user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`url` LIKE \'%https://my-server.org%\')');
  }


  /** @test */
  public function websiteLikeHttpsDoubleDotSlashSlashPlusQuotePommeShouldAnswersOnlyUserWithSlashesPlusAndQuote()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->like('website', '%https://my-server.org/pomme+d\'api%');

    $this->assertEquals(['user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`website` LIKE \'%https://my-server.org/pomme+d\'api%\')');
  }


  /** @test */
  public function urlMatchHttpsDoubleDotSlashSlashShouldAnswersOnlyUserWithSlashes()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('website'))
              ->against_or('https://my-server.org/pomme+d\'api.php'));

    $this->assertEquals(['user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`website`) AGAINST(\'https://my-server.org/pomme+d\'api.php\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withIsNullPrefsShouldAnswersUser_100_Only()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->is_null('prefs');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`prefs` IS null)');
  }


  /** @test */
  public function withIsNotNullPrefsShouldAnswersUsers_101_and_102()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->not_is_null('prefs');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`prefs` IS NOT null)');
  }


  /** @test */
  public function withNotInInviteAdministrateurShouldAnswerUser_100_Only()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->not_in('level', ['invite', 'administrateur']);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`level` NOT IN (\'invite\', \'administrateur\'))');
  }


  /** @test */
  public function withClauseLimitOneInIntAndStringShouldReturnOnlyUser_100()
  {
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit(1)
                        ->fetchAll());

    $this->assertSql('SELECT `users`.* FROM `users` LIMIT 1');
  }


  /** @test */
  public function withEqualLengthOnColumnLevelShouldAnswerUser_102_Only()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->eq(Storm_Query_Key::length('level'), 6);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) = 6)');
  }


  /** @test */
  public function withNotEqualLengthOnColumnLevelShouldAnswerUser_100_101()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->not_eq(Storm_Query_Key::length('level'), 6);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(101)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) != 6)');
  }


  /** @test */
  public function withGreaterLengthOnColumnLevelShouldAnswerUser_100_Only()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->gt(Storm_Query_Key::length('level'), 14);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) > 14)');
  }


  /** @test */
  public function withGreaterEqualLengthOnColumnLevelShouldAnswerUser_100_101()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->gt_eq(Storm_Query_Key::length('level'), 14);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(101)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) >= 14)');
  }


  /** @test */
  public function withLesserLengthOnColumnLevelShouldAnswerUser_102_Only()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->lt(Storm_Query_Key::length('level'), 14);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) < 14)');
  }


  /** @test */
  public function withLesserEqualLengthOnColumnLevelShouldAnswerUser_101_102()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->lt_eq(Storm_Query_Key::length('level'), 14);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) <= 14)');
  }


  /** @test */
  public function withEqualLeftWidth4OnColumnDateShouldAnswerUser_101_Only()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->eq(Storm_Query_Key::left('date', 4), '2020');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) = \'2020\')');
  }


  /** @test */
  public function withNotEqualLeftWidth4OnColumnDateShouldAnswerUser_100_102()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->not_eq(Storm_Query_Key::left('date', 4), '2020');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) != \'2020\')');
  }


  /** @test */
  public function withGreaterLeftWidth4OnColumnDateShouldAnswerUser_100_Only()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->gt(Storm_Query_Key::left('date', 4), '2021');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) > \'2021\')');
  }


  /** @test */
  public function withGreaterEqualLeftColumnDateShouldAnswerUser_100_102()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->gt_eq(Storm_Query_Key::left('date', 4), '2021');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) >= \'2021\')');
  }


  /** @test */
  public function withLesserLeftColumnDateShouldAnswerUser_101_Only()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->lt(Storm_Query_Key::left('date', 4), '2021');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) < \'2021\')');
  }


  /** @test */
  public function withLesserEqualLeftColumnDateShouldAnswerUser_101_102()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->lt_eq(Storm_Query_Key::left('date', 4), '2021');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) <= \'2021\')');
  }


  /** @test */
  public function withClauseLimitPageShouldPaginate()
  {
    // LIMIT 0
    $this->assertEquals([],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(0, 0)
                        ->fetchAll());
    // LIMIT 1
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(0, 1)
                        ->fetchAll());
    // LIMIT 0
    $this->assertEquals([],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(1, 0)
                        ->fetchAll());
    // LIMIT 2
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(101)],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(0, 2)
                        ->fetchAll());
    // LIMIT 2
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(101)],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(1, 2)
                        ->fetchAll());
    // LIMIT 2 OFFSET 2
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(102)],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(2, 2)
                        ->fetchAll());
    // LIMIT 5 OFFSET 10
    $this->assertEquals([],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(3, 5)
                        ->fetchAll());

    $this->assertSqlCount('SELECT `users`.* FROM `users` LIMIT 2', 2);
    $this->assertSqlCount('SELECT `users`.* FROM `users` LIMIT 0', 2);
    $this->assertSqlCount('SELECT `users`.* FROM `users` LIMIT 1', 1);
    $this->assertSqlCount('SELECT `users`.* FROM `users` LIMIT 2 OFFSET 2', 1);
    $this->assertSqlCount('SELECT `users`.* FROM `users` LIMIT 5 OFFSET 10', 1);
  }


  /** @test */
  public function withClauseLimitPageAndAttributeShouldWork()
  {
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(102)],
                        Storm_Test_LoaderQueryUser::query()
                        ->start('foo', 'premier')
                        ->limit_page(1, 2)
                        ->fetchAll());

    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`foo` LIKE \'premier%\') LIMIT 2');
  }


  /** @test */
  public function queriesObserverShouldDisplaySqlAndCallCount()
  {
    foreach (range(1, 2) as $range)
      Storm_Test_LoaderQueryUser::query()
        ->eq('id', 100)
        ->fetchAll();

    foreach (range(1, 5) as $range)
      Storm_Test_LoaderQueryUser::query()
        ->gt('id', 100)
        ->fetchAll();

    Storm_Test_LoaderQueryUser::query()
      ->eq('id', 100)
      ->fetchAll();

    foreach (range(1, 4) as $range)
      Storm_Test_LoaderQueryUser::query()
        ->in('id', [100, 102])
        ->gt('id', 100)
        ->fetchAll();

    $this->assertSqlCount('SELECT `users`.* FROM `users` WHERE (`users`.`id` = 100)',
                          3);
    $this->assertSqlCount('SELECT `users`.* FROM `users` WHERE (`users`.`id` > 100)',
                          5);
    $this->assertSqlCount('SELECT `users`.* FROM `users` WHERE (`users`.`id` IN (100, 102) AND `users`.`id` > 100)',
                          4);
  }


  /** @test */
  public function whereMultipleCallShouldAssemble_2_Queries()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->gt('id', 100);

    $this->assertEquals(['user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());

    $query->in('id', [101, 1000]);

    $this->assertEquals(['user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());

    $this->assertSqlEquals(['SELECT `users`.* FROM `users` WHERE (`users`.`id` > 100)',
                            'SELECT `users`.* FROM `users` WHERE (`users`.`id` > 100 AND `users`.`id` IN (101, 1000))']);
  }


  /** @test */
  public function withEmptyInShouldIgnoreClauseIn()
  {
    Storm_Test_LoaderQueryUser::query()
      ->in('id', [])
      ->fetchAll();

    $this->assertSqlEquals(["SELECT `users`.* FROM `users`"]);
  }


  /** @test */
  public function withEmptyNotInShouldIgnoreClauseIn()
  {
    Storm_Test_LoaderQueryUser::query()
      ->not_in('id', [])
      ->fetchAll();

    $this->assertSqlEquals(["SELECT `users`.* FROM `users`"]);
  }


  /** @test */
  public function inWithOnlyOneValueInShouldTransformToClauseEqual()
  {
    Storm_Test_LoaderQueryUser::query()
      ->in('id', [100])
      ->fetchAll();

    $this->assertSqlEquals(["SELECT `users`.* FROM `users` WHERE (`users`.`id` = 100)"]);
  }


  /** @test */
  public function notInWithOnlyOneValueInShouldTransformToClauseEqual()
  {
    Storm_Test_LoaderQueryUser::query()
      ->not_in('id', [100])
      ->fetchAll();

    $this->assertSqlEquals(["SELECT `users`.* FROM `users` WHERE (`users`.`id` != 100)"]);
  }
}
