<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2024 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
*/

require_once __DIR__ . '/Fixtures.php';

class Storm_Test_Query_LoaderVolatileExistsSubQueriesTest extends Storm_Test_ModelTestCase
{

  protected array $_storm_scopes = ['LoaderVolatileExistsSubQueriesTest'];

  public function setUp(): void
  {
    parent::setUp();

    Storm_Model_Loader::defaultToVolatile();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name 1',
                    'actor_id' => 1,
                    'age' => 21]);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'name' => 'User name 2',
                    'actor_id' => 2,
                    'age' => 22]);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 3,
                    'name' => 'User name 3',
                    'actor_id' => 3,
                    'age' => 23]);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 2,
                    'id_actor' => 2,
                    'label' => 'Actor 2',
                    'age' => 22]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 3,
                    'id_actor' => 3,
                    'label' => 'Actor 3',
                    'age' => 20]);
  }


  public function tearDown(): void
  {
    Storm_Model_Loader::defaultToDb();
    Storm_Model_Loader::resetCache();
    parent::tearDown();
  }


  /** @test */
  public function withExistsSubQueriesReturnOnlyUser_Name3()
  {
    $results = Storm_Test_LoaderQueryUser::query()
      ->exists(Storm_Test_LoaderQueryActor::subExists()
               ->on_eq('actor_id', 'id_actor')
               ->on_gt('age', 'age'))
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['id' => 3,
                         'name' => 'User name 3',
                         'actor_id' => 3,
                         'age' => 23], $results[0]->toArray());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (EXISTS (SELECT \'x\' FROM `actors` WHERE (`users`.`actor_id` = `actors`.`id_actor` AND `users`.`age` > `actors`.`age`)))');
  }


  /** @test */
  public function withExistsSubQueriesReturnOnlyUser_Name2()
  {
    $results = Storm_Test_LoaderQueryUser::query()
      ->exists(Storm_Test_LoaderQueryActor::subExists()
               ->on_gt_eq('actor_id', 'id_actor')
               ->on_lt_eq('age', 'age'))
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['id' => 2,
                         'name' => 'User name 2',
                         'actor_id' => 2,
                         'age' => 22], $results[0]->toArray());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (EXISTS (SELECT \'x\' FROM `actors` WHERE (`users`.`actor_id` >= `actors`.`id_actor` AND `users`.`age` <= `actors`.`age`)))');
  }


  /** @test */
  public function withNotExistsSubQueriesWithTwoConditionReturnUser_Name2_Name3()
  {
    $results = Storm_Test_LoaderQueryUser::query()
      ->not_exists(Storm_Test_LoaderQueryActor::subExists()
                   ->on_not_eq('actor_id', 'id_actor')
                   ->on_lt('age', 'age'))
      ->fetchAll();

    $this->assertCount(2, $results);
    $this->assertEquals(['id' => 2,
                         'name' => 'User name 2',
                         'actor_id' => 2,
                         'age' => 22], $results[0]->toArray());
    $this->assertEquals(['id' => 3,
                         'name' => 'User name 3',
                         'actor_id' => 3,
                         'age' => 23], $results[1]->toArray());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (NOT EXISTS (SELECT \'x\' FROM `actors` WHERE (`users`.`actor_id` != `actors`.`id_actor` AND `users`.`age` < `actors`.`age`)))');
  }


  /** @test */
  public function withNotExistsSubQueriesWithOneConditionAndOneCriteriaReturnUser_Name1_Name2()
  {
    $results = Storm_Test_LoaderQueryUser::query()
      ->not_exists(Storm_Test_LoaderQueryActor::subExists()
                   ->on_eq('actor_id', 'id_actor')
                   ->lt('age', 22))
      ->fetchAll();

    $this->assertCount(2, $results);
    $this->assertEquals(['id' => 1,
                         'name' => 'User name 1',
                         'actor_id' => 1,
                         'age' => 21], $results[0]->toArray());
    $this->assertEquals(['id' => 2,
                         'name' => 'User name 2',
                         'actor_id' => 2,
                         'age' => 22], $results[1]->toArray());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (NOT EXISTS (SELECT \'x\' FROM `actors` WHERE (`users`.`actor_id` = `actors`.`id_actor` AND `actors`.`age` < 22)))');
  }


  /** @test */
  public function withExistsInCriteriaOrShouldHaveSql()
  {
    Storm_Test_LoaderQueryUser::query()
      ->or((new Storm_Query_Criteria)
           ->eq('age', 22)
           ->exists(Storm_Test_LoaderQueryActor::subExists()
                    ->on_eq('actor_id', 'id_actor')
                    ->eq('age', 20)))
      ->fetchAll();

    $this->assertSql('SELECT `users`.* FROM `users` WHERE ((`users`.`age` = 22 OR EXISTS (SELECT \'x\' FROM `actors` WHERE (`users`.`actor_id` = `actors`.`id_actor` AND `actors`.`age` = 20))))');
  }


  /** @test */
  public function existsWithAndConditionOnParentAndChildShouldReturn_Id_3()
  {
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 4,
                    'id_actor' => 3,
                    'label' => 'Actor 4',
                    'age' => 25]);

    $row_results = Storm_Test_LoaderQueryUser::query('tab1')
      ->exists(Storm_Test_LoaderQueryActor::subExists('tab2')
               ->on_eq('actor_id', 'id_actor')
               ->and_sub(['tab1' => (new Storm_Query_Criteria)
                          ->gt('age', 22),
                          'tab2' => (new Storm_Query_Criteria)
                          ->in('age', [22, 25])]))
      ->fetchAll();

    $this->assertEquals([['id' => 3,
                          'name' => 'User name 3',
                          'actor_id' => 3,
                          'age' => 23]],
                        array_map(fn($row) => $row->attributesToArray(),
                                  $row_results));

    $this->assertSql("SELECT `tab1`.* FROM `users` AS `tab1` WHERE (EXISTS (SELECT 'x' FROM `actors` AS `tab2` WHERE (`tab1`.`actor_id` = `tab2`.`id_actor` AND (`tab1`.`age` > 22 AND `tab2`.`age` IN (22, 25)))))");
  }
}
