<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2024 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
*/

require_once __DIR__ . '/Fixtures.php';

class Storm_Test_Query_LoaderGroupWithMinMaxTest extends Storm_Test_ModelTestCase
{

  protected array $_storm_scopes = ['LoaderGroupWithMinMaxTest'];

  public function setUp(): void
  {
    parent::setUp();

    Storm_Model_Loader::defaultToVolatile();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 100,
                    'login' => 'user_admin',
                    'date' => '2022-01-01',
                    'count' => 102,
                    'group' => 'group1',
                    'field' => 'field1']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 101,
                    'login' => 'user_administrateur',
                    'date' => '2020-01-01',
                    'count' => 15000,
                    'group' => 'group1',
                    'field' => 'field1']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 102,
                    'login' => 'user_invite',
                    'date' => '2025-01-06',
                    'count' => 2401,
                    'group' => 'group2',
                    'field' => 'field2']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 103,
                    'login' => 'user_bib',
                    'date' => '2000-12-01',
                    'count' => 98312,
                    'group' => 'group2',
                    'field' => 'field3']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 104,
                    'login' => 'user_portail',
                    'date' => '2021-01-01',
                    'count' => 1000,
                    'group' => 'group1',
                    'field' => 'field4']);
  }


  public function tearDown(): void
  {
    Storm_Model_Loader::defaultToDb();
    Storm_Model_Loader::resetCache();
    parent::tearDown();
  }


  /** @test */
  public function groupByGroupAndSelectMaxDateWithHavingMaxDateShouldAnswers_100()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select(['id', 'group', 'max' => Storm_Query_Key::max('date')])
      ->group('group')
      ->having_gt('max', '2018-01-01')
      ->having_lt('max', '2024-01-01')
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['id' => 100,
                          'group' => 'group1',
                          'max' => '2022-01-01']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql("SELECT `users`.`id`, `users`.`group`, MAX(`users`.`date`) AS `max` FROM `users` GROUP BY `users`.`group` HAVING (`max` > '2018-01-01') AND (`max` < '2024-01-01')");
  }


  /** @test */
  public function groupByGroupAndFieldShouldAnswersUserId_100_102_103_104()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select(['id', 'group', 'field'])
      ->group('group')
      ->group('field')
      ->fetchAll();

    $this->assertCount(4, $row_results);
    $this->assertEquals([['id' => 100,
                          'group' => 'group1',
                          'field' => 'field1'],
                         ['id' => 102,
                          'group' => 'group2',
                          'field' => 'field2'],
                         ['id' => 103,
                          'group' => 'group2',
                          'field' => 'field3'],
                         ['id' => 104,
                          'group' => 'group1',
                          'field' => 'field4']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT `users`.`id`, `users`.`group`, `users`.`field` FROM `users` GROUP BY `users`.`group`, `users`.`field`');
  }


  /** @test */
  public function selectUsersGroupByGroupWithSelectMinOnDateShouldAnswersUserAdministrateurAndUserBib()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::min('date'), 'login', 'group'])
      ->group('group')
      ->fetchAll();

    $this->assertCount(2, $row_results);
    $this->assertEquals([['date' => '2020-01-01',
                          'login' => 'user_administrateur',
                          'group' => 'group1'],
                         ['date' => '2000-12-01',
                          'login' => 'user_bib',
                          'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT MIN(`users`.`date`), `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group`');
  }


  /** @test */
  public function selectUsersGroupByGroupWithSelectMaxOnCountShouldAnswersUserAdministrateurAndUserBib()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::max('count'), 'login', 'group'])
      ->group('group')
      ->fetchAll();

    $this->assertCount(2, $row_results);
    $this->assertEquals([['count' => 15000,
                          'login' => 'user_administrateur',
                          'group' => 'group1'],
                         ['count' => 98312,
                          'login' => 'user_bib',
                          'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT MAX(`users`.`count`), `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group`');
  }


  /** @test */
  public function selectUsersGroupByGroupWithSelectSumOnCountShouldAnswersUserAdminAndUserInvite()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::sum('count'), 'login', 'group'])
      ->group('group')
      ->fetchAll();

    $this->assertCount(2, $row_results);
    $this->assertEquals([['count' => 16102,
                          'login' => 'user_admin',
                          'group' => 'group1'],
                         ['count' => 100713,
                          'login' => 'user_invite',
                          'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT SUM(`users`.`count`), `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group`');
  }


  /** @test */
  public function selectUsersGroupByGroupWithSelectCountShouldAnswersUserAdminAndUserInvite()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select(['count_users' => Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->fetchAll();

    $this->assertCount(2, $row_results);
    $this->assertEquals([['count_users' => 3,
                          'login' => 'user_admin',
                          'group' => 'group1'],
                         ['count_users' => 2,
                          'login' => 'user_invite',
                          'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_users`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group`');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_GT_2_ShouldAnswersUserAdmin()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_gt('count_by', 2)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['count_by' => 3,
                          'login' => 'user_admin',
                          'group' => 'group1']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` > 2)');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_GT_EQ_3_ShouldAnswersUserAdmin()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_gt_eq('count_by', 3)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['count_by' => 3,
                          'login' => 'user_admin',
                          'group' => 'group1']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` >= 3)');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_EQ_3_ShouldAnswersUserAdmin()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_eq('count_by', 3)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['count_by' => 3,
                          'login' => 'user_admin',
                          'group' => 'group1']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` = 3)');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_NOT_EQ_2_ShouldAnswersUserAdmin()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_not_eq('count_by', 2)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['count_by' => 3,
                          'login' => 'user_admin',
                          'group' => 'group1']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` != 2)');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_LT_3_ShouldAnswersUserInvite()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_lt('count_by', 3)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([1 => ['count_by' => 2,
                               'login' => 'user_invite',
                               'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` < 3)');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_LT_EQ_2_ShouldAnswersUserInvite()
  {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_lt_eq('count_by', 2)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([1 => ['count_by' => 2,
                               'login' => 'user_invite',
                               'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` <= 2)');
  }
}
