<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2024 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
*/

require_once __DIR__ . '/Fixtures.php';

class Storm_Test_Query_LoaderVolatileJoinInnerTest extends Storm_Test_ModelTestCase
{

  protected array $_storm_scopes = ['LoaderVolatileJoinInnerTest'];

  public function setUp(): void
  {
    parent::setUp();

    Storm_Model_Loader::defaultToVolatile();

    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 1,
                    'name' => 'Movie name 1',
                    'date' => '2020']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 2,
                    'name' => 'Movie name 2',
                    'date' => '2022']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 3,
                    'name' => 'Movie name 3',
                    'date' => '2021']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 4,
                    'name' => 'Movie name 4',
                    'date' => '2018']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 5,
                    'name' => 'Movie name 5',
                    'date' => '2019']);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 1,
                    'movie_id' => 3,
                    'name' => 'Actor name 1',
                    'age' => 25]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 2,
                    'movie_id' => 1,
                    'name' => 'Actor name 2',
                    'age' => 52]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 3,
                    'movie_id' => 4,
                    'name' => 'Actor name 3',
                    'age' => 36]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 4,
                    'movie_id' => 3,
                    'name' => 'Actor name 4',
                    'age' => 20]);
  }


  public function tearDown(): void
  {
    Storm_Model_Loader::defaultToDb();
    Storm_Model_Loader::resetCache();
    parent::tearDown();
  }


  /** @test */
  public function joinWithOrConditionOnParentAndChildShouldReturn_Id_3_3_4()
  {
    $row_results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->all_select()
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id'))
      ->or_sub(['tab1' => (new Storm_Query_Criteria)
                ->eq('date', '2021'),
                'tab2' => (new Storm_Query_Criteria)
                ->in('age', [36])])
      ->fetchAll();

    $this->assertEquals([['id' => 3,
                          'name' => 'Movie name 3',
                          'date' => '2021'],
                         ['id' => 3,
                          'name' => 'Movie name 3',
                          'date' => '2021'],
                         ['id' => 4,
                          'name' => 'Movie name 4',
                          'date' => '2018']],
                        array_map(fn($row) => $row->attributesToArray(),
                                  $row_results));

    $this->assertSql('SELECT `tab1`.* FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` WHERE ((`tab1`.`date` = \'2021\' OR `tab2`.`age` = 36))');
  }


  /** @test */
  public function joinWithAnd_OrConditionOnParentAndChildShouldReturn_Id_3()
  {
    $row_results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->select('name')
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id')
              ->select('name'))
      ->and_sub(['tab1' => (new Storm_Query_Criteria)
                 ->eq('date', '2021'),
                 'tab2' => (new Storm_Query_Criteria)
                 ->in('age', [20, 36])])
      ->or_sub(['tab1' => (new Storm_Query_Criteria)
                ->eq('name', 'Movie name 3'),
                'tab2' => (new Storm_Query_Criteria)
                ->not_end('name', 'name 1')])
      ->fetchAll();

    $this->assertEquals([['tab1_name' => 'Movie name 3',
                          'tab2_name' => 'Actor name 4']],
                        array_map(fn($row) => $row->attributesToArray(),
                                  $row_results));

    $this->assertSql('SELECT `tab1`.`name` AS `tab1_name`, `tab2`.`name` AS `tab2_name` FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` WHERE ((`tab1`.`date` = \'2021\' AND `tab2`.`age` IN (20, 36)) AND (`tab1`.`name` = \'Movie name 3\' OR `tab2`.`name` NOT LIKE \'%name 1\'))');
  }


  /** @test */
  public function withInnerJoinShouldReturnIntersectMoviesActorsOnlyOneResult()
  {
    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->select(['name', 'date'])
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->select(['name', 'age'])
              ->on_eq('id', 'movie_id')
              ->gt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $row = $results[0];
    $this->assertCount(1, $results);
    $this->assertEquals('Movie name 1', $row->getTab1Name());
    $this->assertEquals('2020', $row->tab1_date);
    $this->assertEquals('Actor name 2', $row->getTab2Name());
    $this->assertEquals(52, $row->tab2_age);
    $this->assertTrue($row instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab1`.`name` AS `tab1_name`, `tab1`.`date` AS `tab1_date`, `tab2`.`name` AS `tab2_name`, `tab2`.`age` AS `tab2_age` FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function withInnerJoinShouldReturnIntersectDedupMovieAndTwoActors()
  {
    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->select(['name', 'date'])
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->select(['name', 'age'])
              ->on_eq('id', 'movie_id')
              ->lt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(2, $results);
    $this->assertEquals([['tab1_name' => 'Movie name 3',
                          'tab1_date' => '2021',
                          'tab2_name' => 'Actor name 1',
                          'tab2_age' => 25],
                         ['tab1_name' => 'Movie name 3',
                          'tab1_date' => '2021',
                          'tab2_name' => 'Actor name 4',
                          'tab2_age' => 20]],
                        array_map(fn($row) => $row->toArray(),
                                  $results));
    $this->assertTrue($results[0] instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab1`.`name` AS `tab1_name`, `tab1`.`date` AS `tab1_date`, `tab2`.`name` AS `tab2_name`, `tab2`.`age` AS `tab2_age` FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` < 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function withInnerJoinWithoutSelectShouldReturnMovies()
  {
    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->all_select()
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id')
              ->gt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['id' => 1,
                         'name' => 'Movie name 1',
                         'date' => '2020'],
                        $results[0]->toArray());
    $this->assertTrue($results[0] instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab1`.* FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function queryPromoteInnerJoinWithAllSelectShouldReturnMovies()
  {
    $results = Storm_Test_LoaderQueryMovie::query()
      ->promoteToJoin('tab1')
      ->all_select()
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id')
              ->gt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(1, $results);
    $row = reset($results);
    $this->assertEquals(1, $row->getId());
    $this->assertEquals('Movie name 1', $row->getName());
    $this->assertEquals('2020', $row->getDate());

    $this->assertSql('SELECT `tab1`.* FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function queryPromoteSubInnerWithSelectShouldReturnMovies()
  {
    $query = Storm_Test_LoaderQueryActor::query()
      ->select('movie_id')
      ->gt('age', 30)
      ->limit(100)
      ->order('age');

    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->all_select()
      ->inner($query->promoteToInner('tab2')
              ->on_eq('id', 'movie_id'))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['id' => 1,
                         'name' => 'Movie name 1',
                         'date' => '2020',
                         'tab2_movie_id' => 1],
                        $results[0]->toArray());

    $this->assertSql('SELECT `tab1`.*, `tab2`.`movie_id` AS `tab2_movie_id` FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function withInnerJoinWithSelectTab2AllShouldReturnActors()
  {
    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id')
              ->all_select()
              ->gt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['id' => 2,
                         'movie_id' => 1,
                         'name' => 'Actor name 2',
                         'age' => 52],
                        $results[0]->toArray());
    $this->assertTrue($results[0] instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab2`.* FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function withInnerJoinWithSelectNameTab2AndSelectAllTab1ShouldReturnActors()
  {
    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->all_select()
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id')
              ->select('name')
              ->gt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['tab2_name' => 'Actor name 2',
                         'id' => 1,
                         'name' => 'Movie name 1',
                         'date' => '2020'],
                        $results[0]->toArray());
    $this->assertTrue($results[0] instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab1`.*, `tab2`.`name` AS `tab2_name` FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }
}
