<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2024 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
*/

require_once __DIR__ . '/Fixtures.php';

class Storm_Test_Query_LoaderUpdateTest extends Storm_Test_ModelTestCase
{

  protected array $_storm_scopes = ['LoaderUpdateTest'];

  public function setUp(): void
  {
    parent::setUp();

    Storm_Model_Loader::defaultToVolatile();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name 1',
                    'age' => 21,
                    'clean_field' => ' field    to be  cleaned   ']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'name' => 'Users names 2',
                    'age' => 22,
                    'clean_field' => 'field dont need clean']);
  }


  public function tearDown(): void
  {
    Storm_Model_Loader::defaultToDb();
    Storm_Model_Loader::resetCache();
    parent::tearDown();
  }


  /** @test */
  public function withOnlyOneSetShouldReturnUpdateQueryAndUpdateVolatile()
  {
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet('name', 'username 2 updated')
      ->match((new Storm_Query_MatchBoolean('age'))
              ->against_or(22))
      ->updateAll();

    $this->assertEquals(1, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21,
                          'clean_field' => ' field    to be  cleaned   '],
                         ['id' => 2,
                          'name' => 'username 2 updated',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql('UPDATE `users` SET `name` = \'username 2 updated\' WHERE (MATCH(`users`.`age`) AGAINST(\'22\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function queryPromoteUpdateWithOnlyOneSetShouldReturnUpdateQueryAndUpdateVolatile()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->select('age')
      ->order('age')
      ->match((new Storm_Query_MatchBoolean('age'))
              ->against_and(22));
    $update = $query->promoteToUpdate();

    $count = $update->addSet('name', 'username 2 updated')
                    ->updateAll();

    $this->assertEquals(1, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21,
                          'clean_field' => ' field    to be  cleaned   '],
                         ['id' => 2,
                          'name' => 'username 2 updated',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql('UPDATE `users` SET `name` = \'username 2 updated\' WHERE (MATCH(`users`.`age`) AGAINST(\'+22\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withTwoSetsShouldReturnUpdateQueryAndUpdateVolatile()
  {
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet('name', 'username 2 updated')
      ->addSet('age', 35)
      ->eq('age', 22)
      ->updateAll();

    $this->assertEquals(1, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21,
                          'clean_field' => ' field    to be  cleaned   '],
                         ['id' => 2,
                          'name' => 'username 2 updated',
                          'age' => 35,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql('UPDATE `users` SET `name` = \'username 2 updated\', `age` = 35 WHERE (`users`.`age` = 22)');
  }


  /** @test */
  public function withCleanSpacesShouldReturnUpdateQueryAndUpdateVolatile()
  {
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet(Storm_Query_Key::cleanSpaces('clean_field'))
      ->updateAll();

    $this->assertEquals(2, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21,
                          'clean_field' => 'field to be cleaned'],
                         ['id' => 2,
                          'name' => 'Users names 2',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql("UPDATE `users` SET `clean_field` = CLEAN_SPACES(`clean_field`)");
  }


  /** @test */
  public function withConcatShouldReturnUpdateQueryAndUpdateVolatile()
  {
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet(Storm_Query_Key::concat('name', ' with concat'))
      ->updateAll();

    $this->assertEquals(2, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1 with concat',
                          'age' => 21,
                          'clean_field' => ' field    to be  cleaned   '],
                         ['id' => 2,
                          'name' => 'Users names 2 with concat',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql("UPDATE `users` SET `name` = CONCAT(`name`, ' with concat')");
  }


  /** @test */
  public function withRegexReplaceShouldReturnUpdateQueryAndUpdateVolatile()
  {
    Storm_Test_LoaderQueryUser::find(1)->setName('http://url.com')->assertSave();
    Storm_Test_LoaderQueryUser::find(2)->setName('url.com')->assertSave();
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet(Storm_Query_Key::regexpReplace('name', 'http[s]*://.*', 'https://new_url.com'))
      ->updateAll();

    $this->assertEquals(2, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'https://new_url.com',
                          'age' => 21,
                          'clean_field' => ' field    to be  cleaned   '],
                         ['id' => 2,
                          'name' => 'url.com',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql("UPDATE `users` SET `name` = REGEXP_REPLACE(`name`, 'http[s]*://.*', 'https://new_url.com')");
  }


  /** @test */
  public function withRegexReplaceCleanSpacesShouldReturnUpdateQueryAndUpdateVolatile()
  {
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet(Storm_Query_Key::regexpReplace('clean_field', '\\bto\\s+be\\b', 'has been')
               ->addCleanSpaces())
      ->updateAll();

    $this->assertEquals(2, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21,
                          'clean_field' => 'field has been cleaned'],
                         ['id' => 2,
                          'name' => 'Users names 2',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql("UPDATE `users` SET `clean_field` = CLEAN_SPACES(REGEXP_REPLACE(`clean_field`, '\\\\bto\\\\s+be\\\\b', 'has been'))");
  }
}
